import os
import platform
from pathlib import Path

import pandas as pd
import pytest

from advertools.spider import crawl

system = platform.system()

links_columns = {
    "links_url": 14,
    "links_text": 14,
    "links_nofollow": 14,
    "nav_links_url": 3,
    "nav_links_text": 3,
    "header_links_url": 3,
    "header_links_text": 3,
    "footer_links_url": 3,
    "footer_links_text": 3,
}

links_filepath = "tests/data/crawl_testing/test_content.html"
if platform == "Windows":
    links_filepath = links_filepath.replace("/", r"\\")

links_file = Path(links_filepath).absolute()
crawl(links_file.as_uri(), "links_crawl.jl", custom_settings={"ROBOTSTXT_OBEY": False})
crawl_df = pd.read_json("links_crawl.jl", lines=True)
os.remove("links_crawl.jl")

crawl(
    str(links_file.as_uri()),
    "follow_url_params.jl",
    allowed_domains=[str(links_file), "example.com"],
    custom_settings={"ROBOTSTXT_OBEY": False},
    follow_links=True,
)

follow_url_params_df = pd.read_json("follow_url_params.jl", lines=True)
os.remove("follow_url_params.jl")


def test_follow_url_params_followed():
    assert follow_url_params_df["url"].str.contains("?", regex=False).any()


crawl(
    str(links_file.as_uri()),
    "dont_follow_url_params.jl",
    allowed_domains=[str(links_file), "example.com"],
    custom_settings={"ROBOTSTXT_OBEY": False},
    follow_links=True,
    exclude_url_params=True,
)
dont_follow_url_params_df = pd.read_json("dont_follow_url_params.jl", lines=True)


def test_dont_follow_url_params_not_followed():
    assert not dont_follow_url_params_df["url"].str.contains("?", regex=False).all()


os.remove("dont_follow_url_params.jl")


file_path = "tests/data/crawl_testing/duplicate_links.html"
if platform == "Windows":
    file_path = links_filepath.replace("/", r"\\")


dup_links_file = Path(file_path).absolute()
crawl(
    str(dup_links_file.as_uri()),
    "dup_links_crawl.jl",
    custom_settings={"ROBOTSTXT_OBEY": False},
)
dup_crawl_df = pd.read_json("dup_links_crawl.jl", lines=True)
os.remove("dup_links_crawl.jl")


def test_link_columns_all_exist():
    assert set(links_columns).difference(crawl_df.columns.tolist()) == set()


@pytest.mark.parametrize("colname,count", links_columns.items())
def test_links_extracted_at_correct_number(colname, count):
    assert crawl_df[colname].str.split("@@").str.len().values[0] == count


def test_extract_h_tags():
    assert crawl_df["h2"].str.split("@@").str.len().values[0] == 3
    assert crawl_df["h2"].str.split("@@").explode().iloc[1] == ""


def test_all_links_have_nofollow():
    assert (
        crawl_df.filter(regex="nofollow")
        .apply(lambda s: s.str.contains("True"))
        .all()
        .all()
    )


def test_image_tags_available():
    assert [
        col in crawl_df for col in ["img_src", "img_alt", "img_height", "img_width"]
    ]


def test_all_img_attrs_have_same_length():
    assert (
        crawl_df.filter(regex="img_")
        .apply(lambda s: s.str.split("@@").str.len())
        .apply(set, axis=1)[0]
        .__len__()
    ) == 1


def test_img_src_has_abs_path():
    assert crawl_df["img_src"].str.startswith("http").all()


dup_links_test = ["https://example_a.com" for i in range(5)] + ["https://example.com"]

dup_text_test = [
    "Link Text A",
    "Link Text A",
    "Link Text A",
    "Link Text B",
    "Link Text C",
    "Link Other",
]

dup_nf_test = ["True"] + ["False" for i in range(5)]


def test_duplicate_links_counted_propery():
    assert dup_crawl_df["links_url"].str.split("@@")[0] == dup_links_test
    assert dup_crawl_df["links_text"].str.split("@@")[0] == dup_text_test
    assert dup_crawl_df["links_nofollow"].str.split("@@")[0] == dup_nf_test


def test_non_existent_links_are_NA():
    assert "nav_links_url" not in dup_crawl_df
    assert "nav_links_text" not in dup_crawl_df
    assert "header_links_url" not in dup_crawl_df
    assert "footer_links_url" not in dup_crawl_df


broken_links_path = "tests/data/crawl_testing/broken_links.html"
if platform == "Windows":
    broken_links_path = links_filepath.replace("/", r"\\")

broken_links_file = Path(broken_links_path).absolute()

crawl(
    [str(broken_links_file.as_uri()), "wrong_url"],
    "broken_links_crawl.jl",
    follow_links=True,
)


def test_broken_links_are_reported():
    broken_links_df = pd.read_json("broken_links_crawl.jl", lines=True)
    assert "errors" in broken_links_df
    assert "wrong_url" not in broken_links_df["url"]
    os.remove("broken_links_crawl.jl")


def test_crawling_bad_url_directly_is_handled():
    crawl(["wrong_url", "https://example.com"], "bad_url.jl")
    bad_url_df = pd.read_json("bad_url.jl", lines=True)
    assert len(bad_url_df) == 1
    assert bad_url_df["url"][0] == "https://example.com"
    os.remove("bad_url.jl")



import platform
from pathlib import Path
from tempfile import TemporaryDirectory

import pytest
from pandas import read_json

from advertools import crawl_headers


@pytest.fixture(scope="session")
def crawl_dir():
    with TemporaryDirectory() as temp_dir:
        return Path(temp_dir).absolute()


@pytest.fixture(scope="session")
def headers_crawl_df(crawl_dir):
    crawl_headers(
        ["https://adver.tools", "does not exist dot com"],
        str(crawl_dir.joinpath("headers_output.jl")),
        custom_settings={
            "ROBOTSTXT_OBEY": False,
            "DEFAULT_REQUEST_HEADERS": {"Accept-Language": "en"},
        },
    )
    filepath = crawl_dir.joinpath("headers_output.jl")
    if platform.system == "Windows":
        filepath = filepath.replace("/", r"\\")

    df = read_json(filepath, lines=True)
    return df



"""
.. _knowledge_graph:

Import and Analyze Knowledge Graph Results on a Large Scale
===========================================================

If :ref:`analyzing SERPs <serp>` is the first step in understanding your
rankings on search engines, then analyzing the knowledge graph can be thought
of as step zero.

SERP positions for a certain keyword show how each page is ranked in comparison
to all other eligible pages. Knowledge graph scores on the other hand, show the
ranks of the different meanings that a word can take for Google (a person, a
city, a brand, etc.).

.. WARNING:: From `Google's documentation <https://developers.google.com/knowledge-graph>`_:
   This API is not suitable for use as a production-critical service. Your
   product should not form a critical dependence on this API.

It's not clear whether this is from a technical reliability or a content
correctness point of view, but it is what the docs mention. So please keep this
in mind when using it.

Account Setup
*************

In order to be able to send requests, you will need to `create a project
<https://console.developers.google.com/>`_, `set up billing
<https://console.developers.google.com/billing>`_, and `activate the knowledge
graph API <https://console.developers.google.com/apis/library>`_ for your
project. You will then need to `create credentials
<https://console.developers.google.com/apis/credentials>`_ (API Key).
Once you have that, you can use it as your ``key`` parameter when running
requests, as shown below.

How to use Google's Knowledge Graph API
***************************************


What is "google"? Is it a search engine, a company, a brand, a very large
number? What else is it?

And if it is all of those things, what is the relative ranking of each? What
is the source of the information, its URL, images (if any)?

.. code-block:: python

    >>> key = 'YOUR_GOOGLE_DEVELOPER_KEY'
    >>> google =  knowledge_graph(key=key, query='google')
    >>> google
         query	resultScore	                               result.@type                  result.description    result.name
    0	google	     203191	   ['Corporation', 'Organization', 'Thing']                  Technology company	        Google
    1	google	      49462	                       ['WebSite', 'Thing']                              Google         Search
    2	google	      19142	                       ['WebSite', 'Thing']                                 nan          Gmail
    3	google	      13251	              ['Brand', 'WebSite', 'Thing']                             Website	   Google Maps
    4	google	       7549	['WebSite', 'SoftwareApplication', 'Thing']                             Website   Google Drive
    5	google	       6853	                       ['WebSite', 'Thing']                             Website	   Google Play
    6	google	       6543	           ['SoftwareApplication', 'Thing']                         Web browser  Google Chrome
    7	google	       4312	   ['Corporation', 'Organization', 'Thing']  Multinational conglomerate company  Alphabet Inc.
    8	google	       3395	           ['SoftwareApplication', 'Thing']                                 nan Google Account
    9	google	       1306	                                  ['Thing']                                 nan         Google

    >>> google.columns
    Index(['query', 'resultScore', '@type', 'result.@type', 'result.description',
       'result.image.contentUrl', 'result.image.url',
       'result.detailedDescription.articleBody',
       'result.detailedDescription.url', 'result.detailedDescription.license',
       'result.url', 'result.name', 'result.@id', 'query_time'],
      dtype='object')



The above table is a sample response from the :func:`knowledge_graph` function.
Many more columns are available as you can see in the second line above.
We can see that "google" is a company, with a result score of 203,191 and it is
a search engine/website with a result score of 49,462. It is then understood as
an email application, a mapping application, and so on, as you can see in the
`result.name` column.

You can also see that we get the types under which this result falls, in the
`result.@type` column. Multiple types show the type inheritance, and as you can
also see, everything is a "Thing". This is the top element in the type
hierarchy under which everything belongs.

Like the :ref:`Google SERP <serp>` and :ref:`YouTube SERP <serp>`, functions
this funcion works in the same manner, creating, sending, and aggregating the
product of the arguments passed
to it.

For example if you run

>>> knowledge_graph(key=key, query=['google', 'bing'], languages=['en', 'fr', 'de'])

The function will send 2 (queries) x 3 languages = 6 requests.

(google, en), (google, fr), (google, de) , (bing, en), (bing, fr), (bing, de)

This is actually the main value of having this function, because you usually
want a large sample to evaluate certain keywords across languages or types.

Let's check what "seo" and "search engine optimization" mean in different
languages.

>>> seo = knowledge_graph(key=key, query=['seo', 'search engine optimization'], languages=['en', 'es', 'de'])
>>> seo
	query	                        languages	resultScore	    result.name	                     result.@type	                                   result.description
0	search engine optimization	de      	       3587	    Suchmaschinenoptimierung	         ['Thing']                                         nan
1	search engine optimization	de      	        321	    Lokale Suchmaschinenoptimierung	 ['Thing']                                         nan
2	search engine optimization	de      	        252	    Suchmaschinenmarketing	         ['Thing']                                         nan
4	search engine optimization	en      	      71756	    Search engine optimization	         ['Thing']                                         nan
5	search engine optimization	en      	       5056	    Search engine marketing	         ['Thing']                                         nan
6	search engine optimization	en      	        576	    SEOP, Inc.	                         ['Organization', 'Corporation', 'Thing']	   Company
13	seo	                        de      	       3313	    Seoul	                         ['AdministrativeArea', 'Thing', 'City', 'Place']  Hauptstadt von Südkorea
14	seo	                        de      	       1509	    Seo Yea-ji	                         ['Thing', 'Person']	                           Schauspielerin
15	seo	                        de      	        584	    Suchmaschinenoptimierung	         ['Thing']	                                   nan
33	seo	                        es      	       1509	    Seo Ye-ji	                         ['Person', 'Thing']	                           Actriz
34	seo	                        es      	        584	    Posicionamiento en buscadores	 ['Thing']	                                   nan
35	seo	                        es      	        316	    Jin	                                 ['Person', 'Thing']	                           Cantante
53	seo	                        en      	       8760	    Search engine optimization	         ['Thing']	                                   nan
54	seo	                        en      	       3313	    Seoul	                         ['AdministrativeArea', 'Thing', 'City', 'Place']  Capital of South Korea
55	seo	                        en      	       1435	    Sulli	                         ['Thing', 'Person']	                           South Korean actress

>>> seo.columns
Index(['query', 'languages', 'resultScore', '@type', 'result.name',
       'result.@type', 'result.@id', 'result.image.contentUrl',
       'result.image.url', 'result.detailedDescription.license',
       'result.detailedDescription.url',
       'result.detailedDescription.articleBody', 'result.description',
       'result.url', 'query_time'],
      dtype='object')

It's interesting to see how the same word can mean different things in
different contexts.

"""

import logging
from concurrent import futures

import pandas as pd
import requests

from advertools.serp import _dict_product

param_regex = '^query$|^ids$|^languages$|^types$|^prefix$|^limit$'

def knowledge_graph(key, query=None, ids=None, languages=None, types=None,
                    prefix=None, limit=None):
    """Query Google's Knowledge Graph with any combination of parameters.

    Note that Google's documentation states that "This API is not suitable for
    use as a production-critical service." So please keep this in mind.

    :param string key: Your Google developer key.
    :param string query: A literal string to search for in the Knowledge Graph.
    :param string ids: A list of entity IDs to search for in the Knowledge
                       Graph.
    :param string languages: The list of language codes (defined in ISO 639) to
                             run the query with, for instance `en`.
    :param string types: Restricts returned entities to those of the specified
                         types. For example,  you can specify `Person` (as
                         defined in http://schema.org/Person)  to restrict the
                         results to entities representing people. If multiple
                         types are specified,  returned entities will contain
                         one or more of these types.
    :param boolean prefix: Enables prefix (initial substring) match against
                           names and aliases of  entities. For example, a
                           prefix `Jung` will match entities and aliases such
                           as `Jung`, `Jungle`, and `Jung-ho Kang`.
    :param number limit: Limits the number of entities to be returned. Maximum
                         is 500. Default is 20.  Requests with high limits have
                         a higher chance of timing out.

    https://developers.google.com/knowledge-graph/reference/rest/v1
    """
    params = locals()
    base_url = 'https://kgsearch.googleapis.com/v1/entities:search?'
    supplied_params = {k: v for k, v in params.items()
                       if params[k] is not None}
    for p in supplied_params:
        if isinstance(supplied_params[p], (str, int)):
            supplied_params[p] = [supplied_params[p]]

    params_list = _dict_product(supplied_params)
    result_df = pd.DataFrame()

    def single_request(param):
        nonlocal result_df
        resp = requests.get(base_url, params=param)
        param_log = ', '.join([k + '=' + str(v) for k, v in param.items()])
        logging.info(msg='Requesting: ' + param_log)
        df = pd.json_normalize(resp.json(), record_path='itemListElement')
        del param['key']
        param_columns = {k: [v] if df.empty else v
                         for k, v in param.items()}
        df = df.assign(**param_columns)
        result_df = pd.concat([result_df, df], ignore_index=True)

    with futures.ThreadPoolExecutor(max_workers=16) as executor:
        executor.map(single_request, params_list)

    reordered_df = pd.concat([result_df.filter(regex=param_regex),
                              result_df.filter(regex=f'^(?!{param_regex})')],
                             axis=1)
    reordered_df['query_time'] = pd.Timestamp.utcnow()
    return reordered_df



import os
from pathlib import Path
from tempfile import TemporaryDirectory

import pandas as pd
import pytest

from advertools.logs import crawllogs_to_df, logs_to_df


def full_local_path(folder, file):
    tests_dir = f"tests/data/{folder}/"
    return os.path.abspath(tests_dir + file)


@pytest.mark.parametrize(
    "logfile",
    [f for f in os.listdir("tests/data/logs_testing") if f.startswith("crawl")],
)
def test_crawllogs_returns_df(logfile):
    result = crawllogs_to_df(full_local_path("logs_testing", logfile))
    assert isinstance(result, pd.core.frame.DataFrame)


def test_logstodf_raises_on_wrong_output_file():
    with pytest.raises(ValueError):
        logs_to_df("logfile.log", "output.wrong", "errors.txt", "common")


def test_logstodf_general_output_correctness():
    logs_to_df(
        full_local_path("logs_testing", "nginx_access.log"),
        "delete_output.parquet",
        "delete_errors.txt",
        "combined",
    )
    result = pd.read_parquet("delete_output.parquet")
    assert "referer" in result
    assert "user_agent" in result
    assert isinstance(result, pd.DataFrame)
    os.remove("delete_output.parquet")
    os.remove("delete_errors.txt")


def test_logs_raises_for_wrong_encoding():
    with pytest.raises(UnicodeDecodeError):
        logs_to_df(
            full_local_path("logs_testing", "combined_latin1_enc.log"),
            "test_output.parquet",
            "test_errors.txt",
            "combined",
            encoding="utf-8",
        )


def test_logs_parses_with_correct_encoding():
    with TemporaryDirectory() as tmpdir:
        path = Path(tmpdir)
        logs_to_df(
            full_local_path("logs_testing", "combined_latin1_enc.log"),
            str(path / "test_output.parquet"),
            path / "test_errors.txt",
            "combined",
            encoding="latin1",
        )
        result = pd.read_parquet(path / "test_output.parquet")
        assert isinstance(result, pd.DataFrame)


def test_logs_nginx_error():
    with TemporaryDirectory() as tmpdir:
        path = Path(tmpdir)
        logs_to_df(
            full_local_path("logs_testing", "nginx_error.log"),
            str(path / "test_output.parquet"),
            path / "test_errors.txt",
            "nginx_error",
            encoding="latin1",
        )
        result = pd.read_parquet(path / "test_output.parquet")
        assert isinstance(result, pd.DataFrame)


def test_logs_apache_error():
    with TemporaryDirectory() as tmpdir:
        path = Path(tmpdir)
        logs_to_df(
            full_local_path("logs_testing", "apache_error.log"),
            str(path / "test_output.parquet"),
            path / "test_errors.txt",
            "nginx_error",
            encoding="latin1",
        )
        result = pd.read_parquet(path / "test_output.parquet")
        assert isinstance(result, pd.DataFrame)


def test_logs_file_without_errors():
    logs_to_df(
        full_local_path("logs_testing", "combined_no_errors.log"),
        "delete_output.parquet",
        "delete_errors.txt",
        "combined",
    )
    assert not os.path.exists("delete_errors.txt")



"""Digital Marketing productivity and analysis tools."""

__author__ = """Elias Dabbas"""
__email__ = "eliasdabbas@gmail.com"
__version__ = "0.14.1"

from advertools.ad_create import ad_create
from advertools.ad_from_string import ad_from_string
from advertools.emoji import *
from advertools.extract import *
from advertools.header_spider import crawl_headers
from advertools.image_spider import crawl_images, summarize_crawled_imgs
from advertools.knowledge_graph import knowledge_graph
from advertools.kw_generate import *
from advertools.logs import LOG_FIELDS, LOG_FORMATS, crawllogs_to_df, logs_to_df
from advertools.regex import *
from advertools.reverse_dns_lookup import reverse_dns_lookup
from advertools.robotstxt import *
from advertools.sitemaps import sitemap_to_df
from advertools.spider import crawl
from advertools.stopwords import stopwords
from advertools.url_builders import url_utm_ga
from advertools.urlytics import url_to_df
from advertools.word_frequency import word_frequency
from advertools.word_tokenize import word_tokenize

from . import crawlytics, twitter, youtube
from .serp import *



"""
.. _crawl_headers:

🕷 Python Status Code Checker with Response Headers
===================================================

A mini crawler that only makes ``HEAD`` requests to a known list of URLs. It
uses `Scrapy <https://docs.scrapy.org/en/latest>`_ under the hood, which means
you get all its power in a simplified interface for a simple and specific
use-case.

The :func:`crawl_headers` function can be used to make those requests for
various quality assurance and analysis reasons. Since ``HEAD`` requests don't
download the whole page, this makes the crawling super light on servers, and
makes the process very fast.

The function is straight-forward and easy to use, you basically need a list of
URLs and a file path where you want to save the output (in `.jl` format):

.. thebe-button::
    Run this code


.. code-block::
    :class: thebe, thebe-init

    import advertools as adv
    import pandas as pd

    url_list = ['https://advertools.readthedocs.io', 'https://adver.tools',
                'https://www.dashboardom.com', 'https://povertydata.org']
    adv.crawl_headers(url_list, 'output_file.jl')
    headers_df = pd.read_json('output_file.jl', lines=True)

    headers_df


====  ============================================  ===================  ========  ==================  =========================  ==================  =======  ==========  ======  =============================  =====================  =============================  ===========================  ===============================  ===============================================================  =================================  ============================  =================================  ===================  ================  ==============  =================================  ==================  ============================================================================  ===============================  =============================  ====================================  =======================  ========================  ============================  ============================  ==========================================  ===========================  ===================================  ===================================  ==============================  =================================  ============================================  ==============================  ==================  =============================  ============================  =======================================================================================  =====================  ===========================================  ==================
  ..  url                                           crawl_time             status    download_timeout  download_slot                download_latency    depth  protocol      body    resp_headers_content-length  resp_headers_server    resp_headers_date              resp_headers_content-type    resp_headers_content-encoding    request_headers_accept                                           request_headers_accept-language    request_headers_user-agent    request_headers_accept-encoding    resp_headers_vary      redirect_times    redirect_ttl  redirect_urls                        redirect_reasons  resp_headers_x-amz-id-2                                                       resp_headers_x-amz-request-id    resp_headers_last-modified     resp_headers_etag                     resp_headers_x-served    resp_headers_x-backend    resp_headers_x-rtd-project    resp_headers_x-rtd-version    resp_headers_x-rtd-path                     resp_headers_x-rtd-domain    resp_headers_x-rtd-version-method    resp_headers_x-rtd-project-method    resp_headers_referrer-policy    resp_headers_permissions-policy    resp_headers_strict-transport-security        resp_headers_cf-cache-status      resp_headers_age  resp_headers_expires           resp_headers_cache-control    resp_headers_expect-ct                                                                   resp_headers_cf-ray    resp_headers_alt-svc                         resp_headers_via
====  ============================================  ===================  ========  ==================  =========================  ==================  =======  ==========  ======  =============================  =====================  =============================  ===========================  ===============================  ===============================================================  =================================  ============================  =================================  ===================  ================  ==============  =================================  ==================  ============================================================================  ===============================  =============================  ====================================  =======================  ========================  ============================  ============================  ==========================================  ===========================  ===================================  ===================================  ==============================  =================================  ============================================  ==============================  ==================  =============================  ============================  =======================================================================================  =====================  ===========================================  ==================
   0  https://adver.tools                           2022-02-11 02:32:26       200                 180  adver.tools                         0.0270483        0  HTTP/1.1       nan                              0  nginx/1.18.0 (Ubuntu)  Fri, 11 Feb 2022 02:32:26 GMT  text/html; charset=utf-8     gzip                             text/html,application/xhtml+xml,application/xml;q=0.9,...;q=0.8  en                                 advertools/0.13.0.rc2         gzip, deflate                      nan                               nan             nan  nan                                               nan  nan                                                                           nan                              nan                            nan                                   nan                      nan                       nan                           nan                           nan                                         nan                          nan                                  nan                                  nan                             nan                                nan                                           nan                                            nan  nan                            nan                           nan                                                                                      nan                    nan                                          nan
   1  https://povertydata.org                       2022-02-11 02:32:26       200                 180  povertydata.org                     0.06442          0  HTTP/1.1       nan                          13270  nginx/1.18.0 (Ubuntu)  Fri, 11 Feb 2022 02:32:26 GMT  text/html; charset=utf-8     gzip                             text/html,application/xhtml+xml,application/xml;q=0.9,...;q=0.8  en                                 advertools/0.13.0.rc2         gzip, deflate                      Accept-Encoding                   nan             nan  nan                                               nan  nan                                                                           nan                              nan                            nan                                   nan                      nan                       nan                           nan                           nan                                         nan                          nan                                  nan                                  nan                             nan                                nan                                           nan                                            nan  nan                            nan                           nan                                                                                      nan                    nan                                          nan
   2  https://advertools.readthedocs.io/en/master/  2022-02-11 02:32:26       200                 180  advertools.readthedocs.io           0.0271282        0  HTTP/1.1       nan                              0  cloudflare             Fri, 11 Feb 2022 02:32:26 GMT  text/html                    gzip                             text/html,application/xhtml+xml,application/xml;q=0.9,...;q=0.8  en                                 advertools/0.13.0.rc2         gzip, deflate                      Accept-Encoding                     1              19  https://advertools.readthedocs.io                 302  rNKT7MYjJ7hcnSvbnZg9qdqizeFfTx9YtZ3/gwNLj8M99yumuCgdd6YTm/iBMO9hrZTAi/iYl50=  EE0DJX6Z511TGX88                 Thu, 10 Feb 2022 17:04:27 GMT  W/"14c904a172315a4922f4d28948b916c2"  Nginx-Proxito-Sendfile   web-i-0710e93d610dd8c3e   advertools                    master                        /proxito/html/advertools/master/index.html  advertools.readthedocs.io    path                                 subdomain                            no-referrer-when-downgrade      interest-cohort=()                 max-age=31536000; includeSubDomains; preload  HIT                                           1083  Fri, 11 Feb 2022 04:32:26 GMT  public, max-age=7200          max-age=604800, report-uri="https://report-uri.cloudflare.com/cdn-cgi/beacon/expect-ct"  6dba2aae6b424107-PRG   h3=":443"; ma=86400, h3-29=":443"; ma=86400  nan
   3  https://www.dashboardom.com                   2022-02-11 02:32:26       200                 180  www.dashboardom.com                 0.118614         0  HTTP/1.1       nan                          26837  gunicorn/19.9.0        Fri, 11 Feb 2022 02:32:26 GMT  text/html; charset=utf-8     nan                              text/html,application/xhtml+xml,application/xml;q=0.9,...;q=0.8  en                                 advertools/0.13.0.rc2         gzip, deflate                      nan                               nan             nan  nan                                               nan  nan                                                                           nan                              nan                            nan                                   nan                      nan                       nan                           nan                           nan                                         nan                          nan                                  nan                                  nan                             nan                                nan                                           nan                                            nan  nan                            nan                           nan                                                                                      nan                    nan                                          1.1 vegur
====  ============================================  ===================  ========  ==================  =========================  ==================  =======  ==========  ======  =============================  =====================  =============================  ===========================  ===============================  ===============================================================  =================================  ============================  =================================  ===================  ================  ==============  =================================  ==================  ============================================================================  ===============================  =============================  ====================================  =======================  ========================  ============================  ============================  ==========================================  ===========================  ===================================  ===================================  ==============================  =================================  ============================================  ==============================  ==================  =============================  ============================  =======================================================================================  =====================  ===========================================  ==================



Optionally, you can customize the crawling behavior with the optional
``custom_settings`` parameter. Please check the
`crawl strategies <_crawl_strategies>`_ page for tips on how you can do that.

Here are some of the common reasons for using a ``HEAD`` crawler:

* **Checking status codes:** One of the most important maintenance tasks you
  should be doing continuously. It's very easy to set up an automated script
  the checks status codes for a few hundred or thousand URLs on a periodic
  basis. You can easily build some rules and alerts based on the status codes
  you get.
* **Status codes of page elements:** Yes, your page returns a 200 OK status,
  but what about all the elements/components of the page? Images, links
  (internal and external), hreflang, canonical, URLs in metatags, script URLs,
  URLs in various structured data elements like Twitter, OpenGraph, and
  JSON-LD are some of the most important ones to check as well.
* **Getting search engine directives:** Those directives can be set using meta
  tags as well as response headers. This crawler gets all available response
  headers so you can check for search engine-specific ones, like `noindex` for
  example.
* **Getting image sizes:** You might want to crawl a list of image URLs and get
  their meta data. The response header `Content-Length` contains the length of
  the page in bytes. With images, it contains the size of the image. This can
  be an extremely efficient way of analyzing image sizes (and other meta data)
  without having to download those images, which could consume a lot of
  bandwidth. Lookout for the column ``resp_headers_content-length``.
* **Getting image types:** The ``resp_headers_content-type`` gives you an
  indication on the type of content of the page (or image when crawling image
  URLs); `text/html`, `image/jpeg` and `image/png` are some such content types.


"""
import datetime
import json
import subprocess

import pandas as pd
import scrapy
from scrapy import Request, Spider

import advertools as adv
from advertools import __version__ as adv_version
from advertools.spider import MAX_CMD_LENGTH, _split_long_urllist

header_spider_path = adv.__path__[0] + '/header_spider.py'

user_agent = f'advertools/{adv_version}'


class HeadersSpider(Spider):
    name = 'headers_spider'
    custom_settings = {
        'USER_AGENT': user_agent,
        'ROBOTSTXT_OBEY': True,
        'HTTPERROR_ALLOW_ALL': True,
        'AUTOTHROTTLE_ENABLED': True,
        'AUTOTHROTTLE_TARGET_CONCURRENCY': 8,

    }

    def __init__(self, url_list=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.start_urls = json.loads(json.dumps(url_list.split(',')))

    def start_requests(self):
        for url in self.start_urls:
            try:
                yield Request(url, callback=self.parse, errback=self.errback,
                              method='HEAD')
            except Exception as e:
                self.logger.error(repr(e))

    def errback(self, failure):
        if not failure.check(scrapy.exceptions.IgnoreRequest):
            self.logger.error(repr(failure))
            now = datetime.datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')
            yield {'url': failure.request.url,
                   'crawl_time': now,
                   'errors': repr(failure)}

    def parse(self, response):
        now = datetime.datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')
        yield {
            'url': response.url,
            'crawl_time': now,
            'status': response.status,
            **{k: '@@'.join(str(val) for val in v) if isinstance(v, list)
               else v for k, v in response.meta.items()},
            'protocol': response.protocol,
            'body': response.text or None,
            **{'resp_headers_' + k: v
               for k, v in response.headers.to_unicode_dict().items()},
            **{'request_headers_' + k: v
               for k, v in response.request.headers.to_unicode_dict().items()},
        }


def crawl_headers(url_list, output_file,  custom_settings=None):
    """Crawl a list of URLs using the HEAD method.

    This function helps in analyzing a set of URLs by getting status codes,
    download latency, all response headers and a few other meta data about the
    crawled URLs.

    Sine the full page is not downloaded, these requests are very light on
    servers and it is super-fast. You can modify the speed of course through
    various settings.

    Typically status code checking is an on-going task that needs to be done
    and managed. Automated alerts can be easily created based on certain status
    codes. Another interesting piece of the information is the `Content-Length`
    response header. This gives you the size of the response body without
    having to download the whole page. It can also be very interesting with
    image URLs. Downloading all images can really be expensive and time
    consuming. Being able to get image sizes without having to download them
    can help a lot in making decisions about optimizing those images.
    Several other data can be interesting to analyze, depending on what
    response headers you get.


    :param url,list url_list: One or more URLs to crawl.
    :param str output_file: The path to the output of the crawl. Jsonlines only
                            is supported to allow for dynamic values. Make sure
                            your file ends with ".jl", e.g. `output_file.jl`.
    :param dict custom_settings: A dictionary of optional custom settings that
                                 you might want to add to the spider's
                                 functionality. There are over 170 settings for
                                 all kinds of options. For details please
                                 refer to the `spider settings <https://docs.scrapy.org/en/latest/topics/settings.html>`_
                                 documentation.
    :Examples:

    >>> import advertools as adv
    >>> url_list = ['https://exmaple.com/A', 'https://exmaple.com/B',
    ...             'https://exmaple.com/C', 'https://exmaple.com/D',
    ...             'https://exmaple.com/E']

    >>> adv.crawl_headers(url_list, 'output_file.jl')
    >>> import pandas as pd
    >>> crawl_df = pd.read_json('output_file.jl', lines=True)
    """
    if isinstance(url_list, str):
        url_list = [url_list]
    if output_file.rsplit('.')[-1] != 'jl':
        raise ValueError("Please make sure your output_file ends with '.jl'.\n"
                         "For example:\n"
                         f"{output_file.rsplit('.', maxsplit=1)[0]}.jl")
    settings_list = []
    if custom_settings is not None:
        for key, val in custom_settings.items():
            if isinstance(val, dict):
                setting = '='.join([key, json.dumps(val)])
            else:
                setting = '='.join([key, str(val)])
            settings_list.extend(['-s', setting])

    command = ['scrapy', 'runspider', header_spider_path,
               '-a', 'url_list=' + ','.join(url_list),
               '-o', output_file] + settings_list
    if len(','.join(url_list)) > MAX_CMD_LENGTH:
        split_urls = _split_long_urllist(url_list)
        for u_list in split_urls:
            command[4] = 'url_list=' + ','.join(u_list)
            subprocess.run(command)
    else:
        subprocess.run(command)



import platform

import pytest

from advertools.header_spider import crawl_headers


def test_crawl_headers_raises_on_wrong_file_extension():
    with pytest.raises(ValueError):
        crawl_headers("https://example.com", "myfile.wrong")


@pytest.mark.parametrize("column", ["url", "crawl_time", "status"])
@pytest.mark.skipif(platform.system() == "Windows", reason="Skip if on Windows")
def test_crawl_headers_returns_df(headers_crawl_df, column):
    assert column in headers_crawl_df



from secrets import token_hex
from tempfile import TemporaryDirectory

import numpy as np
import pandas as pd
import pytest

from advertools.urlytics import url_to_df

domain = "http://example.com"
domain_path = "https://example.com/path"
path_rel = "/path_rel"
path_rel_noslash = "no_slash_no_nothing"
domain_query = "https://www.example.com?one=1&two=2"
domain_query_rel = "/?one=1&two=2"
port = "https://www.example.com:80"
fragment = "https://example.com/#fragment"
fragment_rel = "/#fragment_rel"
full = "ftp://example.com:20/cat/sub_cat?one=10&three=30#frag_2"

ordered_q_param_urls = [
    "https://example.com?a=a&b=b&c=c",
    "https://example.com?b=b&c=c",
    "https://example.com?c=c",
    "https://example.com?a=a&c=c",
]


def test_urltodf_raises_on_wrong_file_ext():
    with pytest.raises(ValueError):
        url_to_df(["https://example.com"], output_file="output.wrong")


def test_urltodf_convert_str_tolist():
    result = url_to_df("https://www.example.com")
    assert isinstance(result, pd.DataFrame)


def test_path_rel_noslash():
    result = url_to_df(path_rel_noslash)
    assert pd.isna(result["scheme"][0])
    assert pd.isna(result["netloc"][0])


def test_abs_and_rel():
    result = url_to_df([domain, path_rel])
    assert "dir_1" in result
    assert len(result) == 2
    assert result["scheme"].iloc[-1] is None
    assert result["query"].isna().all()
    assert result["fragment"].isna().all()


def test_domainpath_fragrel_full():
    result = url_to_df([domain_path, fragment_rel, full])
    assert len(result) == 3
    assert "dir_2" in result
    query_set = {"query_one", "query_three"}
    assert query_set.intersection(result.columns) == query_set


def test_no_path_has_no_last_dir():
    result = url_to_df(domain_query)
    assert "last_dir" not in result


def test_all():
    result = url_to_df(
        [
            domain,
            domain_path,
            path_rel,
            domain_query,
            domain_query_rel,
            port,
            fragment,
            fragment_rel,
            full,
        ]
    )
    assert len(result) == 9
    assert "port" in result
    assert "hostname" in result
    assert "last_dir" in result


def test_query_params_are_ordered_by_fullness():
    result = url_to_df(ordered_q_param_urls)
    query_df = result.filter(regex="^query_")
    sorted_q_params = query_df.notna().mean().sort_values(ascending=False).index
    assert (query_df.columns == sorted_q_params).all()


def test_urltodf_produces_outputfile():
    with TemporaryDirectory() as tmpdir:
        url_to_df(
            ["https://example.com/one/two", "https://example.com/one/two"],
            output_file=f"{tmpdir}/output.parquet",
        )
        df = pd.read_parquet(f"{tmpdir}/output.parquet")
        assert isinstance(df, pd.DataFrame)


def test_urltodf_preserves_order_of_supplied_urls():
    with TemporaryDirectory() as tmpdir:
        urls = [f"https://example.com/one/two/{token_hex(16)}" for i in range(2300)]
        url_to_df(urls, output_file=f"{tmpdir}/output.parquet")
        df = pd.read_parquet(f"{tmpdir}/output.parquet")
        assert df["url"].eq(urls).all()



import json
import re
import subprocess
from urllib.parse import urlsplit

import pandas as pd
from scrapy import Field, Item, Request, Spider
from scrapy.pipelines.images import ImagesPipeline

import advertools as adv

image_spider_path = adv.__path__[0] + '/image_spider.py'

user_agent = f'advertools/{adv.__version__}'


class ImgItem(Item):
    image_urls = Field()
    images = Field()
    image_location = Field()


class AdvImagesPipeline(ImagesPipeline):
    def file_path(self, request, response=None, info=None, *, item=None):
        img_url = request.url
        return urlsplit(img_url).path.split('/')[-1]


class ImageSpider(Spider):
    name = 'image_spider'
    include_img_regex = None
    custom_settings = {
        'USER_AGENT': user_agent,
        'ROBOTSTXT_OBEY': True,
        'HTTPERROR_ALLOW_ALL': True,
        'ITEM_PIPELINES': {AdvImagesPipeline: 1},
        'AUTOTHROTTLE_ENABLED': True,
        'AUTOTHROTTLE_TARGET_CONCURRENCY': 8,
    }

    def __init__(self, start_urls, include_img_regex=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.start_urls = json.loads(json.dumps(start_urls.split(',')))
        if include_img_regex is not None:
            self.include_img_regex = include_img_regex

    def start_requests(self):
        for url in self.start_urls:
            yield Request(url, callback=self.parse)

    def parse(self, response):
        img_item = ImgItem()
        img_src = response.xpath('//img/@src').getall()
        if self.include_img_regex is not None:
            img_src = [response.urljoin(src) for src in img_src
                       if re.findall(self.include_img_regex, src)]
        else:
            img_src = [response.urljoin(src) for src in img_src]
        img_item['image_urls'] = img_src
        img_item['image_location'] = response.request.url
        yield img_item


def crawl_images(
    start_urls,
    output_dir,
    min_width=0,
    min_height=0,
    include_img_regex=None,
    custom_settings=None
):
    """Download all images available on start_urls and save them to output_dir.

    THIS FUNCTION IS STILL EXPERIMENTAL. Expect many changes.

    Parameters
    ----------

    start_urls : list
      A list of URLs from which you want to download available images.
    output_dir : str
      The directory where you want the images to be saved.
    min_width : int
      The minimum width in pixels for an image to be downloaded.
    min_height : int
      The minimum height in pixels for an image to be downloaded.
    include_img_regex : str
      A regular expression to select image src URLs. Use this to restrict image
      files that match this regex.
    custom_settings : dict
      Additional settings to customize the crawling behaviour.
    """
    settings_list = []
    if custom_settings is not None:
        for key, val in custom_settings.items():
            if isinstance(val, dict):
                setting = '='.join([key, json.dumps(val)])
            else:
                setting = '='.join([key, str(val)])
            settings_list.extend(['-s', setting])

    command = [
        'scrapy', 'runspider', image_spider_path,
        '-a', 'start_urls=' + ','.join(start_urls),
        '-s', 'IMAGES_STORE=' + output_dir,
        '-s', 'IMAGES_MIN_HEIGHT=' + str(min_height),
        '-s', 'IMAGES_MIN_WIDTH=' + str(min_width),
        '-o', output_dir + '/image_summary.jl'
        ] + settings_list
    if include_img_regex is not None:
        command += ['-a', 'include_img_regex=' + include_img_regex]
    subprocess.run(command)


def summarize_crawled_imgs(image_dir):
    """Provide a DataFrame of image locations and image URLs resulting from
    crawl_images.

    Running the crawl_images function create a summary CSV file of the
    downloaded images. This function parses that file and provides a two-column
    DataFrame:

    - image_location: The URL from which the images was downloaded from.
    - image_urls: The URL of the image file tha was downloaded.

    Parameters
    ----------
    image_dir : str
      The path to the directory that you provided to crawl_images
    """
    df = pd.read_json(image_dir.rstrip('/') + '/image_summary.jl', lines=True)
    return df[['image_location', 'image_urls']].explode('image_urls')



"""
.. _sitemaps:

Download, Parse, and Analyze XML Sitemaps
=========================================

One of the fastest and easiest ways to get insights on a website's content is
to simply download its XML sitemap(s).

Sitemaps are also important SEO tools as they reveal a lot of information about
the website, and help search engines in indexing those pages. You might want to
run an SEO audit and check if the URLs in the sitemap properly correspond to
the actual URLs of the site, so this would be an easy way to get them.

Sitemaps basically contain a log of publishing activity, and if they have rich
URLs then you can do some good analysis on their content over time as well.

The :func:`sitemap_to_df` function is very simple to use, and only requires the
URL of a sitemap, a sitemap index, or even a robots.txt file. It goes through
the sitemap(s) and returns a DataFrame containing all the tags and their
information.

*  `loc`: The location of the URLs of hte sitemaps.
*  `lastmod`: The datetime of the date when each URL was last modified, if
   available.
*  `sitemap`: The URL of the sitemap from which the URL on this row was
   retreived.
*  `etag`: The entity tag of the response header, if provided.
*  `sitemap_last_modified`: The datetime when the sitemap file was last
   modified, if provided.
*  `sitemap_size_mb`: The size of the sitemap in mega bytes
   (1MB = 1,024 x 1,024 bytes)
*  `download_date`: The datetime when the sitemap was downloaded.

Sitemap Index
-------------
Large websites typically have a sitmeapindex file, which contains links to all
other regular sitemaps that belong to the site. The :func:`sitemap_to_df`
function retreives all sub-sitemaps recursively by default.
In some cases, especially with very large sites, it might be better to first
get the sitemap index, explore its structure, and then decide which sitemaps
you want to get, or if you want them all. Even with smaller websites, it still
might be interesting to get the index only and see how it is structured.

This behavior can be modified by the ``recursive`` parameter, which is set to
`True` by default. Set it to `False` if you want only the index file.

Another interesting thing you might want to do is to provide a robots.txt URL,
and set `recursive=False` to get all available sitemap index files.

>>> sitemap_to_df('https://example.com/robots.txt', recursive=False)

Let's now go through a quick example of what can be done with sitemaps. We can
start by getting one of the BBC's sitemaps.

Regular XML Sitemaps
--------------------

.. thebe-button::
    Run this code

.. code-block::
    :class: thebe, thebe-init

    import advertools as adv
    
    bbc_sitemap = adv.sitemap_to_df('https://www.bbc.com/sitemaps/https-sitemap-com-archive-1.xml')
    bbc_sitemap.head(10)

====  =============================================================================  =========================  ============================================================  ================================  =========================  =================  ================================
  ..  loc                                                                            lastmod                    sitemap                                                       etag                              sitemap_last_modified        sitemap_size_mb  download_date
====  =============================================================================  =========================  ============================================================  ================================  =========================  =================  ================================
   0  https://www.bbc.com/arabic/middleeast/2009/06/090620_as_iraq_explosion_tc2     2009-06-20 14:10:48+00:00  https://www.bbc.com/sitemaps/https-sitemap-com-archive-1.xml  e7e15811c65f406f89f89fe10aef29f5  2021-11-05 20:52:56+00:00            7.63124  2022-02-12 01:37:39.461037+00:00
   1  https://www.bbc.com/arabic/middleeast/2009/06/090620_iraq_blast_tc2            2009-06-20 21:07:43+00:00  https://www.bbc.com/sitemaps/https-sitemap-com-archive-1.xml  e7e15811c65f406f89f89fe10aef29f5  2021-11-05 20:52:56+00:00            7.63124  2022-02-12 01:37:39.461037+00:00
   2  https://www.bbc.com/arabic/business/2009/06/090622_me_worldbank_tc2            2009-06-22 12:41:48+00:00  https://www.bbc.com/sitemaps/https-sitemap-com-archive-1.xml  e7e15811c65f406f89f89fe10aef29f5  2021-11-05 20:52:56+00:00            7.63124  2022-02-12 01:37:39.461037+00:00
   3  https://www.bbc.com/arabic/multimedia/2009/06/090624_me_inpictures_brazil_tc2  2009-06-24 15:27:24+00:00  https://www.bbc.com/sitemaps/https-sitemap-com-archive-1.xml  e7e15811c65f406f89f89fe10aef29f5  2021-11-05 20:52:56+00:00            7.63124  2022-02-12 01:37:39.461037+00:00
   4  https://www.bbc.com/arabic/business/2009/06/090618_tomtest                     2009-06-18 15:32:54+00:00  https://www.bbc.com/sitemaps/https-sitemap-com-archive-1.xml  e7e15811c65f406f89f89fe10aef29f5  2021-11-05 20:52:56+00:00            7.63124  2022-02-12 01:37:39.461037+00:00
   5  https://www.bbc.com/arabic/multimedia/2009/06/090625_sf_tamim_verdict_tc2      2009-06-25 09:46:39+00:00  https://www.bbc.com/sitemaps/https-sitemap-com-archive-1.xml  e7e15811c65f406f89f89fe10aef29f5  2021-11-05 20:52:56+00:00            7.63124  2022-02-12 01:37:39.461037+00:00
   6  https://www.bbc.com/arabic/middleeast/2009/06/090623_iz_cairo_russia_tc2       2009-06-23 13:10:56+00:00  https://www.bbc.com/sitemaps/https-sitemap-com-archive-1.xml  e7e15811c65f406f89f89fe10aef29f5  2021-11-05 20:52:56+00:00            7.63124  2022-02-12 01:37:39.461037+00:00
   7  https://www.bbc.com/arabic/sports/2009/06/090622_me_egypt_us_tc2               2009-06-22 15:37:07+00:00  https://www.bbc.com/sitemaps/https-sitemap-com-archive-1.xml  e7e15811c65f406f89f89fe10aef29f5  2021-11-05 20:52:56+00:00            7.63124  2022-02-12 01:37:39.461037+00:00
   8  https://www.bbc.com/arabic/sports/2009/06/090624_mz_wimbledon_tc2              2009-06-24 13:57:18+00:00  https://www.bbc.com/sitemaps/https-sitemap-com-archive-1.xml  e7e15811c65f406f89f89fe10aef29f5  2021-11-05 20:52:56+00:00            7.63124  2022-02-12 01:37:39.461037+00:00
   9  https://www.bbc.com/arabic/worldnews/2009/06/090623_mz_leaders_lifespan_tc2    2009-06-23 13:24:23+00:00  https://www.bbc.com/sitemaps/https-sitemap-com-archive-1.xml  e7e15811c65f406f89f89fe10aef29f5  2021-11-05 20:52:56+00:00            7.63124  2022-02-12 01:37:39.461037+00:00
====  =============================================================================  =========================  ============================================================  ================================  =========================  =================  ================================

.. thebe-button::
    Run this code

.. code-block::
    :class: thebe, thebe-init

    print(bbc_sitemap.shape)
    print(bbc_sitemap.dtypes)
    
.. code-block::

    (49999, 7)

    loc                                   object
    lastmod                  datetime64[ns, UTC]
    sitemap                               object
    etag                                  object
    sitemap_last_modified    datetime64[ns, UTC]
    sitemap_size_mb                      float64
    download_date            datetime64[ns, UTC]
    dtype: object

Since ``lastmod`` is a ``datetime`` object, we can easily use it for various
time-related operations.
Here we look at how many articles have been published (last modified) per year.

.. thebe-button::
    Run this code

.. code-block::
    :class: thebe, thebe-init

    bbc_sitemap.set_index('lastmod').resample('A')['loc'].count()

.. code-block::

    lastmod
    2008-12-31 00:00:00+00:00     2287
    2009-12-31 00:00:00+00:00    47603
    2010-12-31 00:00:00+00:00        0
    2011-12-31 00:00:00+00:00        0
    2012-12-31 00:00:00+00:00        0
    2013-12-31 00:00:00+00:00        0
    2014-12-31 00:00:00+00:00        0
    2015-12-31 00:00:00+00:00        0
    2016-12-31 00:00:00+00:00        0
    2017-12-31 00:00:00+00:00        0
    2018-12-31 00:00:00+00:00        0
    2019-12-31 00:00:00+00:00       99
    2020-12-31 00:00:00+00:00       10
    Freq: A-DEC, Name: loc, dtype: int64

As the majority are in 2009 with a few in other years, it seems these were
later updated, but we would have to check to verify (in this special case BBC's
URLs contain date information, which can be compared to ``lastmod`` to check if
there is a difference between them).

We can take a look at a sample of the URLs to get the URL template that they
use.

.. thebe-button::
    Run this code

.. code-block::
    :class: thebe, thebe-init

    bbc_sitemap['loc'].sample(10).tolist()

.. code-block::

    ['https://www.bbc.com/russian/rolling_news/2009/06/090628_rn_pakistani_soldiries_ambush',
    'https://www.bbc.com/urdu/pakistan/2009/04/090421_mqm_speaks_rza',
    'https://www.bbc.com/arabic/middleeast/2009/07/090723_ae_silwan_tc2',
    'https://www.bbc.com/portuguese/noticias/2009/07/090729_iraquerefenbritsfn',
    'https://www.bbc.com/portuguese/noticias/2009/06/090623_egitomilitaresfn',
    'https://www.bbc.com/portuguese/noticias/2009/03/090302_gazaconferenciaml',
    'https://www.bbc.com/portuguese/noticias/2009/07/090715_hillary_iran_cq',
    'https://www.bbc.com/vietnamese/culture/2009/04/090409_machienhuu_revisiting',
    'https://www.bbc.com/portuguese/noticias/2009/05/090524_paquistaoupdateg',
    'https://www.bbc.com/arabic/worldnews/2009/06/090629_om_pakistan_report_tc2']

It seems the pattern is

    **https://www.bbc.com/{language}/{topic}/{YYYY}/{MM}/{YYMMDD_article_title}**

This is quite a rich structure, full of useful information. We can
:ref:`analyze the URL structure <urlytics>` using the ``url_to_df`` function:

.. thebe-button::
    Run this code

.. code-block::
    :class: thebe, thebe-init

    url_df = adv.url_to_df(bbc_sitemap['loc'])
    url_df

=====  =============================================================================  ========  ===========  ==========================================================  =======  ==========  ==========  ==========  =======  =======  ===============================  =======  =======  ===============================
   ..  url                                                                            scheme    netloc       path                                                        query    fragment    dir_1       dir_2         dir_3    dir_4  dir_5                              dir_6    dir_7  last_dir
=====  =============================================================================  ========  ===========  ==========================================================  =======  ==========  ==========  ==========  =======  =======  ===============================  =======  =======  ===============================
    0  https://www.bbc.com/arabic/middleeast/2009/06/090620_as_iraq_explosion_tc2     https     www.bbc.com  /arabic/middleeast/2009/06/090620_as_iraq_explosion_tc2                          arabic      middleeast     2009       06  090620_as_iraq_explosion_tc2         nan      nan  090620_as_iraq_explosion_tc2
    1  https://www.bbc.com/arabic/middleeast/2009/06/090620_iraq_blast_tc2            https     www.bbc.com  /arabic/middleeast/2009/06/090620_iraq_blast_tc2                                 arabic      middleeast     2009       06  090620_iraq_blast_tc2                nan      nan  090620_iraq_blast_tc2
    2  https://www.bbc.com/arabic/business/2009/06/090622_me_worldbank_tc2            https     www.bbc.com  /arabic/business/2009/06/090622_me_worldbank_tc2                                 arabic      business       2009       06  090622_me_worldbank_tc2              nan      nan  090622_me_worldbank_tc2
    3  https://www.bbc.com/arabic/multimedia/2009/06/090624_me_inpictures_brazil_tc2  https     www.bbc.com  /arabic/multimedia/2009/06/090624_me_inpictures_brazil_tc2                       arabic      multimedia     2009       06  090624_me_inpictures_brazil_tc2      nan      nan  090624_me_inpictures_brazil_tc2
    4  https://www.bbc.com/arabic/business/2009/06/090618_tomtest                     https     www.bbc.com  /arabic/business/2009/06/090618_tomtest                                          arabic      business       2009       06  090618_tomtest                       nan      nan  090618_tomtest
49994  https://www.bbc.com/vietnamese/world/2009/08/090831_dalailamataiwan            https     www.bbc.com  /vietnamese/world/2009/08/090831_dalailamataiwan                                 vietnamese  world          2009       08  090831_dalailamataiwan               nan      nan  090831_dalailamataiwan
49995  https://www.bbc.com/vietnamese/world/2009/09/090901_putin_regret_pact          https     www.bbc.com  /vietnamese/world/2009/09/090901_putin_regret_pact                               vietnamese  world          2009       09  090901_putin_regret_pact             nan      nan  090901_putin_regret_pact
49996  https://www.bbc.com/vietnamese/culture/2009/09/090901_tiananmen_movie          https     www.bbc.com  /vietnamese/culture/2009/09/090901_tiananmen_movie                               vietnamese  culture        2009       09  090901_tiananmen_movie               nan      nan  090901_tiananmen_movie
49997  https://www.bbc.com/vietnamese/pictures/2009/08/090830_ugc_ddh_sand            https     www.bbc.com  /vietnamese/pictures/2009/08/090830_ugc_ddh_sand                                 vietnamese  pictures       2009       08  090830_ugc_ddh_sand                  nan      nan  090830_ugc_ddh_sand
49998  https://www.bbc.com/vietnamese/business/2009/09/090901_japecontask             https     www.bbc.com  /vietnamese/business/2009/09/090901_japecontask                                  vietnamese  business       2009       09  090901_japecontask                   nan      nan  090901_japecontask
=====  =============================================================================  ========  ===========  ==========================================================  =======  ==========  ==========  ==========  =======  =======  ===============================  =======  =======  ===============================

It seems that the ``dir_1`` is where they have the language information, so we
can easily count how many articles they have per language:

.. thebe-button::
    Run this code

.. code-block::
    :class: thebe, thebe-init

    url_df['dir_1'].value_counts()

.. code-block::

    russian       14022
    persian       10968
    portuguese     5403
    urdu           5068
    mundo          5065
    vietnamese     3561
    arabic         2984
    hindi          1677
    turkce          706
    ukchina         545
    Name: dir_1, dtype: int64

We can also get a subset of articles written in a certain language, and see how
many articles they publish per month, week, year, etc.

.. thebe-button::
    Run this code

.. code-block::
    :class: thebe, thebe-init

    (bbc_sitemap[bbc_sitemap['loc']
     .str.contains('/russian/')]
     .set_index('lastmod')
     .resample('M')['loc'].count())

.. code-block::

    lastmod
    2009-04-30 00:00:00+00:00    1506
    2009-05-31 00:00:00+00:00    2910
    2009-06-30 00:00:00+00:00    3021
    2009-07-31 00:00:00+00:00    3250
    2009-08-31 00:00:00+00:00    2769
                                 ...
    2019-09-30 00:00:00+00:00       8
    2019-10-31 00:00:00+00:00      17
    2019-11-30 00:00:00+00:00      11
    2019-12-31 00:00:00+00:00      24
    2020-01-31 00:00:00+00:00       6
    Freq: M, Name: loc, Length: 130, dtype: int64

The topic or category of the article seems to be in ``dir_2`` for which we can
do the same and count the values.

.. thebe-button::
    Run this code

.. code-block::
    :class: thebe, thebe-init

    url_df['dir_2'].value_counts()[:20]

.. code-block::

    rolling_news        9044
    world               5050
    noticias            4224
    iran                3682
    pakistan            2103
    afghanistan         1959
    multimedia          1657
    internacional       1555
    sport               1350
    international       1293
    india               1285
    america_latina      1274
    business            1204
    cultura_sociedad     913
    middleeast           874
    worldnews            872
    russia               841
    radio                769
    science              755
    football             674
    Name: dir_2, dtype: int64

There is much more you can do, and a lot depends on the URL structure, which
you have to explore and run the right operation.

For example, we can use the ``last_dir`` column which contains the slugs
of the articles, replace underscores with spaces, split, concatenate all, put
in a ``pd.Series`` and count the values. This way we see how many times each
word occurred in an article. The same code can also be run after filtering for
articles in a particular language to get a more meaningful list of words.

.. thebe-button::
    Run this code

.. code-block::
    :class: thebe, thebe-init

    url_df['last_dir'].str.split('_').str[1:].explode().value_counts()[:20]

.. code-block::

    rn          8808
    tc2         3153
    iran        1534
    video        973
    obama        882
    us           862
    china        815
    ir88         727
    russia       683
    si           640
    np           638
    afghan       632
    ka           565
    an           556
    iraq         554
    pakistan     547
    nh           533
    cq           520
    zs           510
    ra           491
    Name: last_dir, dtype: int64

This was a quick overview and data preparation for a sample sitemap. Once you
are familiar with the sitemap's structure, you can more easily start analyzing
the content.

.. note::

    There is a bug currently with tags that contain multiple values in
    sitemaps. If an image column in a news sitemap contains multiple images,
    only the last one is retreived. The same applies for any other sitemap that
    has a tag with multiple values.


News Sitemaps
-------------

.. thebe-button::
    Run this code

.. code-block::
    :class: thebe, thebe-init

    nyt_news = adv.sitemap_to_df('https://www.nytimes.com/sitemaps/new/news.xml.gz')
    print(nyt_news.shape)
    # (5085, 16)
    nyt_news

====  =======================================================================================================  =========================  ======  ==================  ==================  ======================  =======================  ====================================================================================  ========================================================================================================================  =======  ==================================================================================================================================================================================  ==================================================  ================================  =========================  =================  ================================
  ..  loc                                                                                                      lastmod                    news    news_publication    publication_name    publication_language    news_publication_date    news_title                                                                            news_keywords                                                                                                             image    image_loc                                                                                                                                                                           sitemap                                             etag                              sitemap_last_modified        sitemap_size_mb  download_date
====  =======================================================================================================  =========================  ======  ==================  ==================  ======================  =======================  ====================================================================================  ========================================================================================================================  =======  ==================================================================================================================================================================================  ==================================================  ================================  =========================  =================  ================================
   0  https://www.nytimes.com/interactive/2021/us/ottawa-ohio-covid-cases.html                                 2022-02-12 00:00:00+00:00                              The New York Times  en                      2021-01-27T17:00:00Z     Ottawa County, Ohio Covid Case and Exposure Risk Tracker                              Coronavirus (2019-nCoV), States (US), Deaths (Fatalities), United States, Disease Rates                                            https://static01.nyt.com/images/2020/03/29/us/ohio-coronavirus-cases-promo-1585539358901/ohio-coronavirus-cases-promo-1585539358901-articleLarge-v274.png                           https://www.nytimes.com/sitemaps/new/news-6.xml.gz  0cff645fbb74c21791568b78a888967d  2022-02-12 20:17:31+00:00          0.0774069  2022-02-12 20:18:39.744247+00:00
   1  https://www.nytimes.com/interactive/2021/us/hopewell-virginia-covid-cases.html                           2022-02-12 00:00:00+00:00                              The New York Times  en                      2021-01-27T17:00:00Z     Hopewell, Virginia Covid Case and Exposure Risk Tracker                               Coronavirus (2019-nCoV), States (US), Deaths (Fatalities), United States, Disease Rates                                            https://static01.nyt.com/images/2020/03/29/us/virginia-coronavirus-cases-promo-1585539536519/virginia-coronavirus-cases-promo-1585539536519-articleLarge-v271.png                   https://www.nytimes.com/sitemaps/new/news-6.xml.gz  0cff645fbb74c21791568b78a888967d  2022-02-12 20:17:31+00:00          0.0774069  2022-02-12 20:18:39.744247+00:00
   2  https://www.nytimes.com/interactive/2021/us/box-butte-nebraska-covid-cases.html                          2022-02-12 00:00:00+00:00                              The New York Times  en                      2021-01-27T17:00:00Z     Box Butte County, Nebraska Covid Case and Exposure Risk Tracker                       Coronavirus (2019-nCoV), States (US), Deaths (Fatalities), United States, Disease Rates                                            https://static01.nyt.com/images/2020/03/29/us/nebraska-coronavirus-cases-promo-1585539237156/nebraska-coronavirus-cases-promo-1585539237156-articleLarge-v281.png                   https://www.nytimes.com/sitemaps/new/news-6.xml.gz  0cff645fbb74c21791568b78a888967d  2022-02-12 20:17:31+00:00          0.0774069  2022-02-12 20:18:39.744247+00:00
   3  https://www.nytimes.com/interactive/2021/us/stearns-minnesota-covid-cases.html                           2022-02-12 00:00:00+00:00                              The New York Times  en                      2021-01-27T17:00:00Z     Stearns County, Minnesota Covid Case and Exposure Risk Tracker                        Coronavirus (2019-nCoV), States (US), Deaths (Fatalities), United States, Disease Rates                                            https://static01.nyt.com/images/2020/03/29/us/minnesota-coronavirus-cases-promo-1585539172701/minnesota-coronavirus-cases-promo-1585539172701-articleLarge-v282.png                 https://www.nytimes.com/sitemaps/new/news-6.xml.gz  0cff645fbb74c21791568b78a888967d  2022-02-12 20:17:31+00:00          0.0774069  2022-02-12 20:18:39.744247+00:00
   4  https://www.nytimes.com/interactive/2021/us/benton-iowa-covid-cases.html                                 2022-02-12 00:00:00+00:00                              The New York Times  en                      2021-01-27T17:00:00Z     Benton County, Iowa Covid Case and Exposure Risk Tracker                              Coronavirus (2019-nCoV), States (US), Deaths (Fatalities), United States, Disease Rates                                            https://static01.nyt.com/images/2020/03/29/us/iowa-coronavirus-cases-promo-1585539039190/iowa-coronavirus-cases-promo-1585539039190-articleLarge-v286.png                           https://www.nytimes.com/sitemaps/new/news-6.xml.gz  0cff645fbb74c21791568b78a888967d  2022-02-12 20:17:31+00:00          0.0774069  2022-02-12 20:18:39.744247+00:00
5080  https://www.nytimes.com/interactive/2021/us/hodgeman-kansas-covid-cases.html                             2022-02-12 00:00:00+00:00                              The New York Times  en                      2021-01-27T17:00:00Z     Hodgeman County, Kansas Covid Case and Exposure Risk Tracker                          Coronavirus (2019-nCoV), States (US), Deaths (Fatalities), United States, Disease Rates                                            https://static01.nyt.com/images/2020/03/29/us/kansas-coronavirus-cases-promo-1585539054298/kansas-coronavirus-cases-promo-1585539054298-articleLarge-v285.png                       https://www.nytimes.com/sitemaps/new/news-2.xml.gz  f53301c8286f9bf59ef297f0232dcfc1  2022-02-12 20:17:31+00:00          0.914107   2022-02-12 20:18:39.995323+00:00
5081  https://www.nytimes.com/interactive/2021/us/miller-georgia-covid-cases.html                              2022-02-12 00:00:00+00:00                              The New York Times  en                      2021-01-27T17:00:00Z     Miller County, Georgia Covid Case and Exposure Risk Tracker                           Coronavirus (2019-nCoV), States (US), Deaths (Fatalities), United States, Disease Rates                                            https://static01.nyt.com/images/2020/03/29/us/georgia-coronavirus-cases-promo-1585538956622/georgia-coronavirus-cases-promo-1585538956622-articleLarge-v290.png                     https://www.nytimes.com/sitemaps/new/news-2.xml.gz  f53301c8286f9bf59ef297f0232dcfc1  2022-02-12 20:17:31+00:00          0.914107   2022-02-12 20:18:39.995323+00:00
5082  https://www.nytimes.com/interactive/2020/11/03/us/elections/results-west-virginia-house-district-1.html  2022-02-12 00:00:00+00:00                              The New York Times  en                      2020-11-03T17:00:00Z     West Virginia First Congressional District Results: David McKinley vs. Natalie Cline  Elections, Presidential Election of 2020, United States, internal-election-open, House of Representatives, West Virginia           https://static01.nyt.com/images/2020/11/03/us/elections/eln-promo-race-west-virginia-house-1WINNER-mckinleyd/eln-promo-race-west-virginia-house-1WINNER-mckinleyd-articleLarge.png  https://www.nytimes.com/sitemaps/new/news-2.xml.gz  f53301c8286f9bf59ef297f0232dcfc1  2022-02-12 20:17:31+00:00          0.914107   2022-02-12 20:18:39.995323+00:00
5083  https://www.nytimes.com/interactive/2020/11/03/us/elections/results-maine-senate.html                    2022-02-12 00:00:00+00:00                              The New York Times  en                      2020-11-03T17:00:00Z     Maine Senate Results: Susan Collins Defeats Sara Gideon                               Elections, Presidential Election of 2020, United States, internal-election-open, Senate, Maine                                     https://static01.nyt.com/images/2020/11/03/us/elections/eln-promo-race-maine-senateWINNER-collinss/eln-promo-race-maine-senateWINNER-collinss-articleLarge.png                      https://www.nytimes.com/sitemaps/new/news-2.xml.gz  f53301c8286f9bf59ef297f0232dcfc1  2022-02-12 20:17:31+00:00          0.914107   2022-02-12 20:18:39.995323+00:00
5084  https://www.nytimes.com/interactive/2021/us/randolph-missouri-covid-cases.html                           2022-02-12 00:00:00+00:00                              The New York Times  en                      2021-01-27T17:00:00Z     Randolph County, Missouri Covid Case and Exposure Risk Tracker                        Coronavirus (2019-nCoV), States (US), Deaths (Fatalities), United States, Disease Rates                                            https://static01.nyt.com/images/2020/03/29/us/missouri-coronavirus-cases-promo-1585539206866/missouri-coronavirus-cases-promo-1585539206866-articleLarge-v282.png                   https://www.nytimes.com/sitemaps/new/news-2.xml.gz  f53301c8286f9bf59ef297f0232dcfc1  2022-02-12 20:17:31+00:00          0.914107   2022-02-12 20:18:39.995323+00:00
====  =======================================================================================================  =========================  ======  ==================  ==================  ======================  =======================  ====================================================================================  ========================================================================================================================  =======  ==================================================================================================================================================================================  ==================================================  ================================  =========================  =================  ================================

Video Sitemaps
--------------

.. thebe-button::
    Run this code

.. code-block::
    :class: thebe, thebe-init

    wired_video = adv.sitemap_to_df('https://www.wired.com/video/sitemap.xml')
    print(wired_video.shape)
    # (2955, 14)
    wired_video

====  ==============================================================================================================  =======  ======================================================================================================================================================================================================================================  ==============================================================================  ==========================================================================================================================================================================================================================================================================================================================================  ========================================================================================================  ================  =========================  =======================  =========  =======================================  ==================================  =================  ================================
  ..  loc                                                                                                             video    video_thumbnail_loc                                                                                                                                                                                                                     video_title                                                                     video_description                                                                                                                                                                                                                                                                                                                           video_content_loc                                                                                           video_duration  video_publication_date       video_expiration_date  lastmod    sitemap                                  etag                                  sitemap_size_mb  download_date
====  ==============================================================================================================  =======  ======================================================================================================================================================================================================================================  ==============================================================================  ==========================================================================================================================================================================================================================================================================================================================================  ========================================================================================================  ================  =========================  =======================  =========  =======================================  ==================================  =================  ================================
   0  https://www.wired.com/video/watch/autocomplete-inverviews-owen-wilson-answers-the-webs-most-searched-questions           http://dwgyu36up6iuz.cloudfront.net/heru80fdn/image/upload/c_fill,d_placeholder_thescene.jpg,fl_progressive,g_face,h_180,q_80,w_320/v1644595412/wired_autocomplete-inverviews-owen-wilson-answers-the-webs-most-searched-questions.jpg  Autocomplete Interview - Owen Wilson Answers The Web’s Most Searched Questions  Owen Wilson takes the WIRED Autocomplete Interview and answers the internet's most searched questions about himself. How did Owen Wilson break his nose? How many movies is he in with Ben Stiller? Is Owen in every Wes Anderson movie? Is he a good skateboarder? Owen answers all these questions and much more!                         http://dp8hsntg6do36.cloudfront.net/62067f085577c277dd9acf42/39687acb-505b-4c69-94f1-afaa7cb5e636low.mp4               645  2022-02-11T17:00:00+00:00                      nan  NaT        https://www.wired.com/video/sitemap.xml  W/90b11f47f8b2ab57cb180cbd3c6f06f9            2.86199  2022-02-12 20:24:55.841851+00:00
   1  https://www.wired.com/video/watch/wired-news-and-science-samsung-s22                                                     http://dwgyu36up6iuz.cloudfront.net/heru80fdn/image/upload/c_fill,d_placeholder_thescene.jpg,fl_progressive,g_face,h_180,q_80,w_320/v1644418652/wired_wired-news-and-science-samsung-s22.jpg                                            Currents - Samsung S22 Ultra Explained in 3 Minutes                             Julian Chokkattu, Reviews Editor for WIRED, walks us through a few of the Samsung S22 Ultra's new features.                                                                                                                                                                                                                                 http://dp8hsntg6do36.cloudfront.net/6203cd7b5577c23d19622259/fe546b9b-a320-4883-9cbd-0d790f23c36dlow.mp4               184  2022-02-10T17:00:00+00:00                      nan  NaT        https://www.wired.com/video/sitemap.xml  W/90b11f47f8b2ab57cb180cbd3c6f06f9            2.86199  2022-02-12 20:24:55.841851+00:00
   2  https://www.wired.com/video/watch/first-look-samsung-galaxy-unpacked-2022                                                http://dwgyu36up6iuz.cloudfront.net/heru80fdn/image/upload/c_fill,d_placeholder_thescene.jpg,fl_progressive,g_face,h_180,q_80,w_320/v1644381627/wired_first-look-samsung-galaxy-unpacked-2022.jpg                                       First Look: Samsung Galaxy Unpacked 2022                                        Samsung has debuted three new smartphones—the Galaxy S22 Ultra, S22+, S22—and three Android tablets in various sizes at Samsung Unpacked 2022. WIRED's Julian Chokkattu takes a look at the newest features.                                                                                                                                http://dp8hsntg6do36.cloudfront.net/620345a15577c23d46622256/d74930cf-11e1-466e-b023-1d9b91664204low.mp4               373  2022-02-09T15:00:00+00:00                      nan  NaT        https://www.wired.com/video/sitemap.xml  W/90b11f47f8b2ab57cb180cbd3c6f06f9            2.86199  2022-02-12 20:24:55.841851+00:00
   3  https://www.wired.com/video/watch/reinventing-with-data                                                                  http://dwgyu36up6iuz.cloudfront.net/heru80fdn/image/upload/c_fill,d_placeholder_thescene.jpg,fl_progressive,g_face,h_180,q_80,w_320/v1642801328/wired_reinventing-with-data.jpg                                                         Reinventing With Data | WIRED Brand Lab                                         Produced by WIRED Brand Lab with AWS | What can the Seattle Seahawks winning strategy teach businesses?                                                                                                                                                                                                                                     http://dp8hsntg6do36.cloudfront.net/619bd9be1d75db41adee6b58/d4889b15-4f34-41b0-b935-0c79465a9793low.mp4               292  2022-02-09T13:00:00+00:00                      nan  NaT        https://www.wired.com/video/sitemap.xml  W/90b11f47f8b2ab57cb180cbd3c6f06f9            2.86199  2022-02-12 20:24:55.841851+00:00
                                                                                                                                                                                                                                                                                                                                                                                                                                                       Swami Sivasubramanian, VP of AI at Amazon Web Services helps us to understand how the Seattle Seahawks are using data and AI to remain a top performing team in the NFL, and how their process of data capture, storage, and machine learning to gain strategic insights is a model for making better business decision across industries.
   4  https://www.wired.com/video/watch/seth-rogen-answers-the-webs-most-searched-questions                                    http://dwgyu36up6iuz.cloudfront.net/heru80fdn/image/upload/c_fill,d_placeholder_thescene.jpg,fl_progressive,g_face,h_180,q_80,w_320/v1644335726/wired_seth-rogen-answers-the-webs-most-searched-questions.jpg                           Autocomplete Interview - Seth Rogen Answers The Web’s Most Searched Questions   "Pam &amp; Tommy" star Seth Rogen takes the WIRED Autocomplete Interview once again and answers the internet's most searched questions about himself. Who does Seth Rogen look like? Does Seth have a podcast? Does he sell pottery? Does he celebrate Christmas? Does he play Call of Duty?                                                http://dp8hsntg6do36.cloudfront.net/6201430a1d75db06ae1f62e8/488ed635-91d0-4281-9e64-34be9bf74f00low.mp4               635  2022-02-08T17:00:00+00:00                      nan  NaT        https://www.wired.com/video/sitemap.xml  W/90b11f47f8b2ab57cb180cbd3c6f06f9            2.86199  2022-02-12 20:24:55.841851+00:00

                                                                                                                                                                                                                                                                                                                                                                                                                                                       Pam &amp; Tommy premieres February 2 on Hulu (finale on March 9)
2950  https://www.wired.com/video/genres/how-to                                                                       nan      nan                                                                                                                                                                                                                                     nan                                                                             nan                                                                                                                                                                                                                                                                                                                                         nan                                                                                                                    nan  nan                                            nan  NaT        https://www.wired.com/video/sitemap.xml  W/90b11f47f8b2ab57cb180cbd3c6f06f9            2.86199  2022-02-12 20:24:55.841851+00:00
2951  https://www.wired.com/video/genres/movies-tv                                                                    nan      nan                                                                                                                                                                                                                                     nan                                                                             nan                                                                                                                                                                                                                                                                                                                                         nan                                                                                                                    nan  nan                                            nan  NaT        https://www.wired.com/video/sitemap.xml  W/90b11f47f8b2ab57cb180cbd3c6f06f9            2.86199  2022-02-12 20:24:55.841851+00:00
2952  https://www.wired.com/video/genres/events                                                                       nan      nan                                                                                                                                                                                                                                     nan                                                                             nan                                                                                                                                                                                                                                                                                                                                         nan                                                                                                                    nan  nan                                            nan  NaT        https://www.wired.com/video/sitemap.xml  W/90b11f47f8b2ab57cb180cbd3c6f06f9            2.86199  2022-02-12 20:24:55.841851+00:00
2953  https://www.wired.com/video/genres/promotion                                                                    nan      nan                                                                                                                                                                                                                                     nan                                                                             nan                                                                                                                                                                                                                                                                                                                                         nan                                                                                                                    nan  nan                                            nan  NaT        https://www.wired.com/video/sitemap.xml  W/90b11f47f8b2ab57cb180cbd3c6f06f9            2.86199  2022-02-12 20:24:55.841851+00:00
2954  https://www.wired.com/video/genres/transportation                                                               nan      nan                                                                                                                                                                                                                                     nan                                                                             nan                                                                                                                                                                                                                                                                                                                                         nan                                                                                                                    nan  nan                                            nan  NaT        https://www.wired.com/video/sitemap.xml  W/90b11f47f8b2ab57cb180cbd3c6f06f9            2.86199  2022-02-12 20:24:55.841851+00:00
====  ==============================================================================================================  =======  ======================================================================================================================================================================================================================================  ==============================================================================  ==========================================================================================================================================================================================================================================================================================================================================  ========================================================================================================  ================  =========================  =======================  =========  =======================================  ==================================  =================  ================================

"""
import logging
from concurrent import futures
from gzip import GzipFile
from urllib.request import Request, urlopen
from xml.etree import ElementTree

import pandas as pd

from advertools import __version__ as version

logging.basicConfig(level=logging.INFO)

headers = {'User-Agent': 'advertools-' + version}


def _sitemaps_from_robotstxt(robots_url):
    sitemaps = []
    robots_page = urlopen(Request(robots_url, headers=headers))
    for line in robots_page.readlines():
        line_split = [s.strip() for s in line.decode().split(':', maxsplit=1)]
        if line_split[0].lower() == 'sitemap':
            sitemaps.append(line_split[1])
    return sitemaps


def _parse_sitemap(root):
    d = dict()
    for node in root:
        for n in node:
            if 'loc' in n.tag:
                d[n.text] = {}

    def parse_xml_node(node, node_url, prefix=''):
        nonlocal d
        keys = []
        for element in node:
            if element.text:
                tag = element.tag.split('}')[-1]
                d[node_url][prefix + tag] = element.text
                keys.append(tag)
                prefix = prefix if tag in keys else ''
            if list(element):
                parse_xml_node(element, node_url, prefix=element.tag.split('}')[-1] + '_')
    for node in root:
        node_url = [n.text for n in node if 'loc' in n.tag][0]
        parse_xml_node(node, node_url=node_url)
    return pd.DataFrame(d.values())


def sitemap_to_df(sitemap_url, max_workers=8, recursive=True):
    """
    Retrieve all URLs and other available tags of a sitemap(s) and put them in
    a DataFrame.

    You can also pass the URL of a sitemap index, or a link to a robots.txt
    file.

    :param url sitemap_url: The URL of a sitemap, either a regular sitemap, a
                            sitemap index, or a link to a robots.txt file.
                            In the case of a sitemap index or robots.txt, the
                            function will go through all the sub sitemaps and
                            retrieve all the included URLs in one DataFrame.
    :param int max_workers: The maximum number of workers to use for threading.
                            The higher the faster, but with high numbers you
                            risk being blocked and/or missing some data as you
                            might appear like an attacker.
    :param bool recursive: Whether or not to follow and import all sub-sitemaps
                           (in case you have a sitemap index), or to only
                           import the given sitemap. This might be useful in
                           case you want to explore what sitemaps are available
                           after which you can decide which ones you are
                           interested in.
    :return sitemap_df: A pandas DataFrame containing all URLs, as well as
                        other tags if available (``lastmod``, ``changefreq``,
                        ``priority``, or others found in news, video, or image
                        sitemaps).
    """
    if sitemap_url.endswith('robots.txt'):
        return pd.concat([sitemap_to_df(sitemap, recursive=recursive)
                          for sitemap in _sitemaps_from_robotstxt(sitemap_url)],
                         ignore_index=True)
    if sitemap_url.endswith('xml.gz'):
        xml_text = urlopen(Request(sitemap_url,
                                   headers={'Accept-Encoding': 'gzip',
                                            'User-Agent': 'advertools-' +
                                                          version}))
        resp_headers = xml_text.getheaders()
        xml_text = GzipFile(fileobj=xml_text)
    else:
        xml_text = urlopen(Request(sitemap_url, headers=headers))
        resp_headers = xml_text.getheaders()
    xml_string = xml_text.read()
    root = ElementTree.fromstring(xml_string)

    sitemap_df = pd.DataFrame()

    if (root.tag.split('}')[-1] == 'sitemapindex') and recursive:
        multi_sitemap_df = pd.DataFrame()
        sitemap_url_list = []
        for elem in root:
            for el in elem:
                if 'loc' in el.tag:
                    if el.text == sitemap_url:
                        error_df = pd.DataFrame({
                            'sitemap': [sitemap_url],
                            'errors': ['WARNING: Sitemap contains a link to itself']
                        })
                        multi_sitemap_df = pd.concat(
                            [multi_sitemap_df, error_df], ignore_index=True)
                    else:
                        sitemap_url_list.append(el.text)
        with futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
            to_do = []
            for sitemap in sitemap_url_list:
                future = executor.submit(sitemap_to_df, sitemap)
                to_do.append(future)
            done_iter = futures.as_completed(to_do)
            for future in done_iter:
                try:
                    multi_sitemap_df = pd.concat(
                        [multi_sitemap_df, future.result()], ignore_index=True)
                except Exception as e:
                    error_df = pd.DataFrame(dict(errors=str(e)),
                                            index=range(1))
                    future_str = hex(id(future))
                    hexes = [hex(id(f)) for f in to_do]
                    index = hexes.index(future_str)
                    error_df['sitemap'] = sitemap_url_list[index]
                    logging.warning(msg=str(e) + ' ' + sitemap_url_list[index])
                    multi_sitemap_df = pd.concat(
                        [multi_sitemap_df, error_df], ignore_index=True)
        return multi_sitemap_df

    else:
        logging.info(msg='Getting ' + sitemap_url)
        elem_df = _parse_sitemap(root)
        sitemap_df = pd.concat([sitemap_df, elem_df], ignore_index=True)
        sitemap_df['sitemap'] = [sitemap_url] if sitemap_df.empty else sitemap_url
    if 'lastmod' in sitemap_df:
        try:
            sitemap_df['lastmod'] = pd.to_datetime(sitemap_df['lastmod'], utc=True)
        except Exception as e:
            pass
    if 'priority' in sitemap_df:
        try:
            sitemap_df['priority'] = sitemap_df['priority'].astype(float)
        except Exception as e:
            pass
    etag_lastmod = {header.lower().replace('-', '_'): val
                    for header, val in resp_headers
                    if header.lower() in ['etag', 'last-modified']}
    sitemap_df = sitemap_df.assign(**etag_lastmod)
    if 'last_modified' in sitemap_df:
        sitemap_df['sitemap_last_modified'] = pd.to_datetime(sitemap_df['last_modified'])
        del sitemap_df['last_modified']
    sitemap_df['sitemap_size_mb'] = len(xml_string) / 1024 / 1024
    sitemap_df['download_date'] = pd.Timestamp.now(tz='UTC')
    return sitemap_df



import os
import random
from itertools import product
from tempfile import TemporaryDirectory, TemporaryFile

import pandas as pd
import pytest

from advertools import crawlytics

test_filepath = "tests/data/crawl_testing/crawlytics.jl"
crawldf = pd.read_json(test_filepath, lines=True)
redirect_df = crawlytics.redirects(crawldf)
link_df = crawlytics.links(crawldf)
link_df_internal = crawlytics.links(crawldf, internal_url_regex="nytimes.com")
image_df = crawlytics.images(crawldf)
rand_columns = [random.choices(crawldf.columns, k=5) for i in range(10)]
regexes = ["img_", "jsonld", "resp_header", r"h\d$"]


def test_redirects_empty_df_redir_urls_isna():
    crawldf_no_redirects = crawldf[crawldf["redirect_urls"].isna()]
    redirect_df = crawlytics.redirects(crawldf_no_redirects)
    assert redirect_df.empty


def test_redirects_empty_df_no_redir_urls_col():
    crawldf = pd.DataFrame()
    redirect_df = crawlytics.redirects(crawldf)
    assert redirect_df.empty


def test_redirects_correct_columns():
    assert set(redirect_df.columns) == {
        "url",
        "status",
        "order",
        "type",
        "download_latency",
        "redirect_times",
    }


def test_redirects_num_redirects_matches_crawldf():
    assert redirect_df.index.nunique() == crawldf["redirect_urls"].dropna().shape[0]


def test_redirects_correct_types():
    assert redirect_df["type"].drop_duplicates().sort_values().tolist() == [
        "crawled",
        "intermediate",
        "requested",
    ]


def test_redirects_correct_redir_times():
    index_counts_df = redirect_df.index.value_counts().sort_index().sub(1)
    redirect_times = (
        redirect_df.reset_index()[["index", "redirect_times"]]
        .drop_duplicates(subset=["index"])
        .set_index("index")
        .squeeze()
    )
    assert index_counts_df.eq(redirect_times).all()


def test_redirects_requested_eq_1():
    assert (
        redirect_df[["order", "type"]].query('type=="requested"')["order"].eq(1).all()
    )


def test_redirects_crawl_status_not_3xx():
    assert (
        redirect_df[["status", "type"]]
        .query('type=="crawled"')["status"]
        .astype(str)
        .str[0]
        .ne("3")
        .all()
    )


def test_redirects_order_monotonic_increasing():
    assert (
        redirect_df.reset_index()
        .groupby("index")["order"]
        .is_monotonic_increasing.all()
    )


def test_links_correct_num_links_preserved():
    index_counts = link_df.index.value_counts().sort_index()
    link_counts = crawldf["links_url"].str.split("@@").str.len().fillna(1)
    assert (index_counts == link_counts).all()


def test_links_same_urls():
    assert set(crawldf["url"]) == set(link_df["url"])


def test_link_urls_same_link_urls():
    assert set(crawldf["links_url"].str.split("@@").explode().dropna()) == set(
        link_df["link"].dropna()
    )


def test_link_text_same_link_text():
    assert set(crawldf["links_text"].str.split("@@").explode().dropna()) == set(
        link_df["text"].dropna()
    )


def test_links_same_nofollows():
    crawl_df_nofollow = [
        eval(str(x)) for x in crawldf["links_nofollow"].str.split("@@").explode()
    ]
    link_df_nofollow = [eval(str(x)) for x in link_df["nofollow"]]
    assert crawl_df_nofollow == link_df_nofollow


def test_links_empty_df_if_no_links_url():
    assert crawlytics.links(crawldf.drop("links_url", axis=1)).empty


def test_links_regex_matches_internal():
    assert (
        link_df_internal["link"]
        .astype(str)
        .str.contains("nytimes.com")
        .eq(link_df_internal["internal"])
        .all()
    )


def test_images_same_num_images():
    crawl_img_counts = (
        crawldf["img_src"].str.split("@@").explode().index.value_counts().sort_index()
    )
    img_df_counts = image_df.index.value_counts().sort_index()
    pd.testing.assert_series_equal(crawl_img_counts, img_df_counts)


def test_links_gets_same_columns():
    assert set(image_df.columns) == set(crawldf.filter(regex="^url$|^img_").columns)


def test_jl_subset_raises_on_no_params():
    with pytest.raises(ValueError):
        crawlytics.jl_subset("filepath.jl")


def test_jl_subset_raises_on_wrong_filepath():
    with pytest.raises(ValueError):
        crawlytics.jl_subset("somewrongfilepath", regex="regex")


@pytest.mark.parametrize("columns", rand_columns)
def test_jl_subset_correct_cols(columns):
    subset_df = crawlytics.jl_subset(test_filepath, columns=columns)
    col_regex = "^" + "$|^".join(columns) + "$"
    assert set(subset_df.columns) == set(subset_df.filter(regex=col_regex).columns)


@pytest.mark.parametrize("regex", regexes)
def test_jl_subset_correct_regex(regex):
    subset_df = crawlytics.jl_subset(test_filepath, regex=regex)
    assert set(subset_df.columns) == set(subset_df.filter(regex=regex).columns)


@pytest.mark.parametrize(["columns", "regex"], product(rand_columns, regexes))
def test_jl_subset_correct_cols_and_regex(columns, regex):
    subset_df = crawlytics.jl_subset(test_filepath, columns=columns, regex=regex)
    col_regex = "^" + "$|^".join(columns) + "$"
    full_regex = "|".join([col_regex, regex])
    assert set(subset_df.columns) == set(subset_df.filter(regex=full_regex).columns)


def test_jl_subset_doesnt_contain_nonexistent_col():
    subset_df = crawlytics.jl_subset(
        test_filepath,
        columns=["title", "url", "doesnt_exist"],
    )
    assert "doesnt_exist" not in subset_df


def test_jl_to_parquet_file_exists():
    with TemporaryDirectory() as tempdir:
        crawlytics.jl_to_parquet(test_filepath, f"{tempdir}/delete.parquet")
        assert os.path.isfile(f"{tempdir}/delete.parquet")


def test_jl_to_parquet_correct_columns():
    with TemporaryDirectory() as tempdir:
        crawlytics.jl_to_parquet(test_filepath, f"{tempdir}/delete.parquet")
        jl_df = pd.read_json(test_filepath, lines=True)
        pq_df = pd.read_parquet(f"{tempdir}/delete.parquet")
        pq_cols = crawlytics.parquet_columns(f"{tempdir}/delete.parquet")
        assert set(jl_df.columns) == set(pq_df.columns)
        assert set(jl_df.columns) == set(pq_cols["column"])



import unittest
from itertools import combinations, permutations

import pandas as pd

from advertools.kw_generate import *


class KeywordTests(unittest.TestCase):
    def test_match_types_are_capitalized(self):
        df = kw_generate(
            ["one", "two"], ["three", "four"], match_types=["eXact", "PHRase", "BRoad"]
        )
        self.assertEqual(set(df["Criterion Type"]), {"Exact", "Phrase", "Broad"})

    def test_raises_error_for_wrong_match_types(self):
        with self.assertRaises(ValueError):
            kw_generate(
                ["one", "two"], ["three", "four"], match_types=["hello", "world"]
            )

    def test_raises_error_for_wrong_max_len(self):
        with self.assertRaises(ValueError):
            kw_generate(["one", "two"], ["three", "four"], max_len=1)

    def test_right_combinatorics_fuction_selected(self):
        df_perm = kw_generate(["one", "two"], ["three", "four"], order_matters=True)
        self.assertTrue({"one three", "three one"}.issubset(set(df_perm["Keyword"])))
        df_comb = kw_generate(["one", "two"], ["three", "four"], order_matters=False)
        self.assertFalse({"one three", "three one"}.issubset(set(df_comb["Keyword"])))

    def test_correct_column_names_used(self):
        headers = ["Campaign", "Ad Group", "Keyword", "Criterion Type", "Labels"]
        df = kw_generate(["one", "two"], ["three", "four"])
        self.assertEqual(list(df.columns.values), headers)

    def test_all_products_in_final_result(self):
        products = {"One", "Two", "Three"}
        df = kw_generate(products, ["four", "five"])
        self.assertEqual(products, set(df["Ad Group"]))

    def test_keyword_lengths_vary_between_2_max_len(self):
        df2 = kw_generate(["one", "two"], ["three", "four"], max_len=2)
        self.assertEqual(set(df2["Keyword"].str.count(" ") + 1), {2})
        df3 = kw_generate(["one", "two"], ["three", "four"], max_len=3)
        self.assertEqual(set(df3["Keyword"].str.count(" ") + 1), {2, 3})
        df4 = kw_generate(["one", "two"], ["three", "four", "five"], max_len=4)
        self.assertEqual(set(df4["Keyword"].str.count(" ") + 1), {2, 3, 4})

    def test_correct_combinations_are_generated(self):
        products = ["one", "two"]
        words = ["three", "four"]
        test_combs = []
        for i in range(2, 4):
            for prod in products:
                for comb in combinations([prod] + words, i):
                    if prod not in comb:
                        continue
                    test_combs.append(" ".join(comb))

        df = kw_generate(products, words, match_types=["Exact"], order_matters=False)
        self.assertEqual(set(test_combs), set(df["Keyword"]))

    def test_correct_permutations_are_generated(self):
        products = ["one", "two"]
        words = ["three", "four"]
        test_combs = []
        for i in range(2, 4):
            for prod in products:
                for perm in permutations([prod] + words, i):
                    if prod not in perm:
                        continue
                    test_combs.append(" ".join(perm))

        df = kw_generate(products, words, match_types=["Exact"], order_matters=True)
        self.assertEqual(set(test_combs), set(df["Keyword"]))

    def test_correct_campaign_name(self):
        df = kw_generate(["one", "two"], ["three"], campaign_name="My_campaign")
        self.assertEqual(df["Campaign"][0], "My_campaign")
        self.assertTrue(len(set(df["Campaign"])) == 1)

    def test_ad_group_names_capitalized(self):
        df = kw_generate(["oNe", "TWO", "THree", "four", "fivE"], ["six"])
        self.assertEqual(
            set(["One", "Two", "Three", "Four", "Five"]), set(df["Ad Group"])
        )

    def test_modified_adds_plus_sign_as_many_as_spaces(self):
        df = kw_generate(["one", "two"], ["three"], match_types=["Modified"])
        self.assertTrue(
            all(df["Keyword"].str.count(r"\+") == df["Keyword"].str.count(" ") + 1)
        )
        self.assertTrue(set(df["Criterion Type"]) == {"Broad"})

    def test_all_match_types_used(self):
        df_exact = kw_generate(["one"], ["two"], match_types=["Exact"])
        df_phrase = kw_generate(["one"], ["two"], match_types=["Phrase"])
        df_broad = kw_generate(["one"], ["two"], match_types=["Broad"])
        df_e_p_b = kw_generate(
            ["one"], ["two"], match_types=["Exact", "Phrase", "Broad", "Modified"]
        )
        self.assertEqual(set(df_exact["Criterion Type"]), {"Exact"})
        self.assertEqual(set(df_phrase["Criterion Type"]), {"Phrase"})
        self.assertEqual(set(df_broad["Criterion Type"]), {"Broad"})
        self.assertEqual(set(df_e_p_b["Criterion Type"]), {"Exact", "Phrase", "Broad"})

    def test_returns_pd_dataframe(self):
        df = kw_generate(["one"], ["two"], order_matters=True)
        self.assertEqual(type(df), pd.core.frame.DataFrame)
        self.assertEqual(df.shape, (6, 5))


words = ["[one]", '"two"', "+three -four", "[five", 'six"', "+seven eight nine]"]


def test_kw_broad():
    result = kw_broad(words)
    assert result == ["one", "two", "three -four", "five", "six", "seven eight nine"]


def test_kw_exact():
    result = kw_exact(words)
    assert result == [
        "[one]",
        "[two]",
        "[three -four]",
        "[five]",
        "[six]",
        "[seven eight nine]",
    ]


def test_exact_contains_brackets():
    result = kw_exact(words)
    assert ["[" in x and "]" in x for x in result]


def test_kw_phrase():
    result = kw_phrase(words)
    assert result == [
        '"one"',
        '"two"',
        '"three -four"',
        '"five"',
        '"six"',
        '"seven eight nine"',
    ]


def test_kw_modified():
    result = kw_modified(words)
    assert result == [
        "+one",
        "+two",
        "+three +-four",
        "+five",
        "+six",
        "+seven +eight +nine",
    ]


def test_kw_neg_broad():
    result = kw_neg_broad(words)
    assert result == [
        "-one",
        "-two",
        "-three -four",
        "-five",
        "-six",
        "-seven eight nine",
    ]


def test_kw_neg_exact():
    result = kw_neg_exact(words)
    assert result == [
        "-[one]",
        "-[two]",
        "-[three -four]",
        "-[five]",
        "-[six]",
        "-[seven eight nine]",
    ]


def test_kw_neg_phrase():
    result = kw_neg_phrase(words)
    assert result == [
        '-"one"',
        '-"two"',
        '-"three -four"',
        '-"five"',
        '-"six"',
        '-"seven eight nine"',
    ]



import pytest

from advertools.spider import _crawl_or_not

url = "https://example.com"


@pytest.mark.parametrize(
    "urlpath", ["", "?one=1", "/one/two/three", "/one/two?hello=yes&1=2"]
)
def test_no_conditions(urlpath):
    result = _crawl_or_not(url + urlpath)
    assert result


def test_exclude_url_params_true():
    result = _crawl_or_not(url + "?1=1", exclude_url_params=True)
    assert not result


def test_reject_right_param():
    result = _crawl_or_not(url + "?one=1", exclude_url_params=["one", "two"])
    assert not result


def test_dont_reject_param():
    result = _crawl_or_not(url + "?hello=1", exclude_url_params=["one", "two"])
    assert result


def test_include_right_param():
    result1 = _crawl_or_not(url + "?one=1&two=2", include_url_params=["one"])
    result2 = _crawl_or_not(url + "?one=1&two=2", include_url_params=["one", "two"])
    assert result1 and result2


def test_include_and_exclude():
    result = _crawl_or_not(
        url + "?one=1&two=2&three=3",
        exclude_url_params=["hello"],
        include_url_params=["one", "four"],
    )
    assert result


def test_exclude_simple_regex():
    result = _crawl_or_not(url, exclude_url_regex="example")
    assert not result


def test_exclude_regex():
    result = _crawl_or_not(url, exclude_url_regex="https:.*ple")
    assert not result


def test_include_simple_regex():
    result = _crawl_or_not(url, include_url_regex="example")
    assert result


def test_include_regex():
    result = _crawl_or_not(url, include_url_regex="https:.*ple")
    assert result


def test_multi_condition():
    result = _crawl_or_not(url, exclude_url_params=True, include_url_regex="example")
    assert result



import os
from advertools.robotstxt import robotstxt_to_df, robotstxt_test
import pandas as pd
import pytest
from .test_sitemaps import full_path

robots_file = full_path('robots.txt')


def test_robotstxt_to_df():
    result = robotstxt_to_df('https://www.media-supermarket.com/robots.txt')
    assert isinstance(result, pd.core.frame.DataFrame)
    assert all(col in result
               for col in ['directive', 'content', 'download_date'])


def test_robtostxt_to_df_handles_list():
    result = robotstxt_to_df([
        'https://www.media-supermarket.com/robots.txt',
        robots_file
    ])
    assert isinstance(result, pd.core.frame.DataFrame)
    assert all(col in result
               for col in ['directive', 'content', 'download_date'])


def test_robotstxt_to_df_saves_single_file():
    robotstxt_to_df('https://www.media-supermarket.com/robots.txt',
                    output_file='robots_output.jl')
    result = pd.read_json('robots_output.jl', lines=True)
    assert isinstance(result, pd.core.frame.DataFrame)
    assert all(col in result
               for col in ['directive', 'content', 'download_date'])
    os.remove('robots_output.jl')


def test_robotstxt_to_df_saves_file_list():
    robotstxt_to_df(['https://www.media-supermarket.com/robots.txt',
                     robots_file],
                    output_file='robots_output.jl')
    result = pd.read_json('robots_output.jl', lines=True)
    assert isinstance(result, pd.core.frame.DataFrame)
    assert all(col in result
               for col in ['directive', 'content', 'download_date'])
    os.remove('robots_output.jl')


def test_robotstxt_to_df_raises_on_wrong_file():
    with pytest.raises(ValueError):
        robotstxt_to_df(robots_file, output_file='wrong_extension.pdf')


def test_robotstxt_to_df_contains_errors():
    result = robotstxt_to_df('wrong_url.html')
    assert 'errors' in result


def test_robotstxt_test():
    user_agents = ['Googlebot', 'Baiduspider', '*']
    urls_to_check = ['/', '/help', 'something.html']
    result = robotstxt_test(robots_file, user_agents, urls_to_check)
    assert isinstance(result, pd.core.frame.DataFrame)
    assert all(col in result for col in
               ['robotstxt_url', 'user_agent', 'url_path', 'can_fetch'])


def test_robotstxt_raises():
    with pytest.raises(ValueError):
        robotstxt_test('http://www.wrong-url.com', '*', '/')


def test_robots_converts_str_to_list():
    result = robotstxt_test('https://www.apple.com/robots.txt', '*', 'hello')
    assert isinstance(result, pd.core.frame.DataFrame)
    assert all(col in result for col in
               ['robotstxt_url', 'user_agent', 'url_path', 'can_fetch'])



"""
After crawling a website, or a bunch of URLs, you mostly likely want to analyze the data
and gain a better undersanding of the website's structure, strategy, and content. You
probably also want to check for technical issues that the site might have.

This module provides a few ready-made functions to help in anayzing crawl data.

.. raw:: html

   <iframe width="560" height="315" src="https://www.youtube.com/embed/rt0LhxNW8GM?si=Pm5v7JKUK5CiS-Lo" title="YouTube video player" frameborder="0" allow="accelerometer; autoplay; clipboard-write; encrypted-media; gyroscope; picture-in-picture; web-share" allowfullscreen></iframe>

|
There are certain columns in the crawl DataFrame that can be analyzed separately and
independently, like page size and status codes. They can of course be analyzed together
with other columns like URL and title to put these columns and their data in context.

There are also groups of columns the can be thought of as describing one full aspect of
a website, yet spread across a few columns. For exmaple:

Analyzing crawled images
------------------------

Every crawled URL typically contains multiple images. Every image in turn has multiple
attributes (src, alt, width, etc.)
The number of images per URL is not the same, and not all images have the same
attributes. So we need a way to unpack all these data points in a tidy (long form)
DataFrame to get an idea of how images (and their attributes) are used and distributed
across the crawled website (URLs).

Once you have read a crawl output file into a DataFrame, you can summarize the images
in this DataFrame as follows:

>>> import advertools as adv
>>> import pandas as pd
>>> crawldf = pd.read_json('path/to/output_file.jl', lines=True)
>>> img_df = adv.crawlytics.images(crawldf)
>>> img_df

====  ===========================================================  ==================================================================================================================================================================================  ================  =============  ================================  ==============  ===========  ============  ============
  ..  url                                                          img_src                                                                                                                                                                             img_alt           img_loading    img_sizes                         img_decoding      img_width    img_height    img_border
====  ===========================================================  ==================================================================================================================================================================================  ================  =============  ================================  ==============  ===========  ============  ============
   0  https://www.nytimes.com/                                     /vi-assets/static-assets/icon-the-morning_144x144-b12a6923b6ad9102b766352261b1a847.webp                                                                                             The Morning Logo                 nan                               nan                     nan           nan           nan
   0  https://www.nytimes.com/                                     /vi-assets/static-assets/icon-the-upshot_144x144-0b1553ff703bbd07ac8fe73e6d215888.webp                                                                                              The Upshot Logo                  nan                               nan                     nan           nan           nan
   0  https://www.nytimes.com/                                     https://static01.nyt.com/images/2017/01/29/podcasts/the-daily-album-art/the-daily-album-art-square320-v5.jpg?quality=75&auto=webp&disable=upscale                                   The Daily Logo                   nan                               nan                     nan           nan           nan
   1  https://www.nytimes.com/newsletters/morning-briefing-europe  https://static.nytimes.com/email-images/NYT-Newsletters-Europe-Icon-500px.jpg                                                                                                       morning briefing  nan            nan                               nan                     nan           nan           nan
   2  https://www.nytimes.com/newsletters/australia-letter         https://static.nytimes.com/email-images/NYT-Newsletters-AustraliaLetter-Icon-500px.jpg                                                                                              australia-letter  nan            nan                               nan                     nan           nan           nan
   3  https://www.nytimes.com/newsletters/the-interpreter          https://static.nytimes.com/email-images/NYT-Newsletters-SONL-TheInterpreter-Icon-500px.jpg                                                                                          the interpreter   nan            nan                               nan                     nan           nan           nan
   4  https://www.nytimes.com/section/world/middleeast             https://static01.nyt.com/images/2024/01/25/multimedia/25israel-1-hbcz/25israel-1-hbcz-thumbWide.jpg?quality=75&auto=webp&disable=upscale                                                              nan            (min-width: 1024px) 205px, 150px  async                   150           100           nan
   4  https://www.nytimes.com/section/world/middleeast             https://static01.nyt.com/images/2024/01/25/multimedia/25israel-hamas-icj-case-explain-wjth/25israel-hamas-icj-case-explain-wjth-thumbWide.jpg?quality=75&auto=webp&disable=upscale                    nan            (min-width: 1024px) 205px, 150px  async                   150           100           nan
   4  https://www.nytimes.com/section/world/middleeast             https://static01.nyt.com/images/2024/01/25/multimedia/25israel-hamas-qatar-israel-ctbv/25israel-hamas-qatar-israel-ctbv-thumbWide.jpg?quality=75&auto=webp&disable=upscale                            nan            (min-width: 1024px) 205px, 150px  async                   150           100           nan
====  ===========================================================  ==================================================================================================================================================================================  ================  =============  ================================  ==============  ===========  ============  ============

As you can see, for every URL we have all available image attributes listed, and in many
cases with those attributes empty, because they were not used for that particular image.
Also note that each image is represented independently on its own row, and mapped to the
URL on which it was found, which can be seen in the first column.
This is why we have the same URL repeated, to represent data for each image. You can
use those URLs to get more data about the URL of interest from the crawl DataFrame.

Let's get a quick overview of the usage of the various image attributes in this
DataFrame. We do this by checking whether a tag is `notna` and get the averages.

>>> img_df.notna().mean().sort_values(ascending=False).to_frame().round(2)

============  ====
url           1
img_src       0.99
img_alt       0.99
img_width     0.86
img_height    0.86
img_srcset    0.25
img_sizes     0.25
img_decoding  0.25
img_loading   0.01
img_border    0
============  ====

We can now see that almost all (99%) of our images have `src` and `alt` attributes.
About 86% have `width` and `height`, and so on. This immediately gives us an overview of
how our images are managed on the site. We can easily estimate the size of the issues if
any, and plan our work accordingly.

Analyzing links in a crawled website
------------------------------------

Another important aspect of any webpage/website is understanding how its pages are
linked, internally and externally.

The ``crawlytics.links`` function gives you a summary of the links, that is similar to
the format of the ``crawlytics.images`` DataFrame.

>>> link_df = adv.crawlytics.links(crawldf, internal_url_regex='nytimes.com')
>>> link_df

====  ===========================================================  ========================================================================  ==================  ==========  ==========
  ..  url                                                          link                                                                      text                nofollow    internal
====  ===========================================================  ========================================================================  ==================  ==========  ==========
   0  https://www.nytimes.com/                                     https://www.nytimes.com/#site-content                                     Skip to content     False       True
   0  https://www.nytimes.com/                                     https://www.nytimes.com/#site-index                                       Skip to site index  False       True
   0  https://www.nytimes.com/                                     https://www.nytimes.com/#after-dfp-ad-top                                 SKIP ADVERTISEMENT  False       True
   1  https://www.nytimes.com/newsletters/morning-briefing-europe  https://www.nytimes.com/newsletters/morning-briefing-europe#site-content  Skip to content     False       True
   1  https://www.nytimes.com/newsletters/morning-briefing-europe  https://www.nytimes.com/newsletters/morning-briefing-europe#site-index    Skip to site index  False       True
   1  https://www.nytimes.com/newsletters/morning-briefing-europe  https://www.nytimes.com/                                                                      False       True
   2  https://www.nytimes.com/newsletters/australia-letter         https://www.nytimes.com/newsletters/australia-letter#site-content         Skip to content     False       True
   2  https://www.nytimes.com/newsletters/australia-letter         https://www.nytimes.com/newsletters/australia-letter#site-index           Skip to site index  False       True
   2  https://www.nytimes.com/newsletters/australia-letter         https://www.nytimes.com/                                                                      False       True
   3  https://www.nytimes.com/newsletters/the-interpreter          https://www.nytimes.com/newsletters/the-interpreter#site-content          Skip to content     False       True
   3  https://www.nytimes.com/newsletters/the-interpreter          https://www.nytimes.com/newsletters/the-interpreter#site-index            Skip to site index  False       True
   3  https://www.nytimes.com/newsletters/the-interpreter          https://www.nytimes.com/                                                                      False       True
   4  https://www.nytimes.com/section/world/middleeast             https://www.nytimes.com/section/world/middleeast#site-content             Skip to content     False       True
   4  https://www.nytimes.com/section/world/middleeast             https://www.nytimes.com/section/world/middleeast#site-index               Skip to site index  False       True
   4  https://www.nytimes.com/section/world/middleeast             https://www.nytimes.com/section/world/middleeast                          Middle East         False       True
====  ===========================================================  ========================================================================  ==================  ==========  ==========

Every link is represented on a separate row, and we have a few attributes for each link,
it's text, whether or not it has a nofollow rel attribute, and optionally whether or not
it is internal. For the optional `internal` parameter you will have to supply a regex to
define what internal really means. You could include certain sub-domains, or even
consider other domains as part of your same property, and thus they would be considered
"internal".

We can now easily count how many links we have per URL, the most frequently used link
text, and so on.

We now take a look at redirects.

Analyzing the redirects of a crawled website
--------------------------------------------

Like images and links, the information about redirects is presented using a group of
columns:

>>> redirect_df = adv.crawlytics.redirects(crawldf)
>>> redirect_df

====  ==================================================================  ========  =======  ============  ==================  ================
  ..  url                                                                   status    order  type            download_latency    redirect_times
====  ==================================================================  ========  =======  ============  ==================  ================
   0  https://nytimes.com                                                      301        1  requested              0.220263                  1
   0  https://www.nytimes.com/                                                 200        2  crawled                0.220263                  1
  26  https://www.nytimes.com/privacy/privacy-policy                           301        1  requested              0.079844                  1
  26  https://help.nytimes.com/hc/en-us/articles/10940941449492                403        2  crawled                0.079844                  1
 105  https://www.nytimes.com/es/privacy/privacy-policy                        301        1  requested              0.0630789                 1
 105  https://help.nytimes.com/hc/en-us/articles/13537530305428                403        2  crawled                0.0630789                 1
 218  https://nytimes.com/spotlight/privacy-project-data-protection            301        1  requested              0.852014                  1
 218  https://www.nytimes.com/spotlight/privacy-project-data-protection        200        2  crawled                0.852014                  1
 225  https://nytimes.com/spotlight/privacy-project-regulation-solutions       301        1  requested              0.732559                  1
 310  http://nytimes.com/by/sahil-chinoy                                       301        1  requested              0.435062                  2
 310  https://nytimes.com/by/sahil-chinoy                                      301        2  intermediate           0.435062                  2
 310  https://www.nytimes.com/by/sahil-chinoy                                  200        3  crawled                0.435062                  2
====  ==================================================================  ========  =======  ============  ==================  ================

Here each redirect is represented using a group of columns, as well as
a group of rows. Columns show attributes of a redirect (status code, the order of the 
URL in the redirect, the type of the URL in the redirect context, download latency in
seconds, and the number of redirects in this specific process).
Since a redirect contains multiple URLs, each one of those URLs is represented on its
own row. You can use the index of this DataFrame to connect a redirect back to the crawl
DataFrame in case you want more context about it.

Let's now see what can be done with large crawl files.

Handling very large crawl files
-------------------------------

Many times you might end up crawling a large website, and the crawl file file might be
as large as your memory (or even more), making it impossible to analyze.

We have several options availablel to us:

  1. Read a subset of columns
  2. Convert the jsonlines file to parquet
  3. Explore the available column names and their respective data types of a parquet
     file

The ``jl_subset`` function only reads the column subset that you want, massively
reducing the memory consumption of our file.
In some cases you only want a small set of columns, you can read the DataFrame with the
columns of interest, write them to a new file, and delete the old large crawl file.

>>> crawl_subset = adv.crawlytics.jl_subset(
...    filepath='/path/to/output_file.jl',
...    columns=[col1, col2, ...],
...    regex=column_regex)

You can use the ``columns`` parameter to specify exactly which columns you want. You can
also use a regular expression to specify a set of columns. Here are some examples of
regular expressions that you might typically want to use:

* "img\_": Get all image columns, including all availabe `<img>` attributes.
* "jsonld\_": Get all JSON-LD columns.
* "resp_headers\_": Response headers.
* "request_headers\_": Request headers.
* "h\\\d": Heading columns, h1..h6.
* "redirect\_": Columns containing redirect information.
* "links\_": Columns containing link information.

An important characteristic of these groups of columns is that you most likely don't
know how many they are, and what they might include, so a regular expression can save a
lot of time.

You can use the `columns` and `regex` parameters together or either one of them on its
own depending on your needs.

Compressing large crawl files
-----------------------------

Another strategy while dealing with large jsonlines (.jl) files is to convert them to
the highly performant `.parquet` format. You simply have to provide the current path to
the .jl file, and provide a path for the desired .parquet file:

>>> adv.crawlytics.jl_to_parquet("output_file.jl", "output_file.parquet")

Now you have a much smaller file on disk, and you can use the full power of parquet to
efficiently read (and filter) columns and rows. Check the`pandas.read_parquet
<https://pandas.pydata.org/docs/reference/api/pandas.read_parquet.html>`_ documentation
for details.


Exploring the columns and data types of parquet files
-----------------------------------------------------

Another simple function gives us a DataFrame of the available columns in a parquet file.
One of the main advantags of using parquet is that you can select which columns you want
to read.

>>> adv.crawlytics.parquet_columns('output_file.parquet') # first 15 columns only

====  ==============  ======
  ..  column          type
====  ==============  ======
   0  url             string
   1  title           string
   2  meta_desc       string
   3  viewport        string
   4  charset         string
   5  h1              string
   6  h2              string
   7  h3              string
   8  canonical       string
   9  alt_href        string
  10  alt_hreflang    string
  11  og:url          string
  12  og:type         string
  13  og:title        string
  14  og:description  string
====  ==============  ======

Check how many columns we have of each type.

>>> adv.crawlytics.parquet_columns('nyt_crawl.parquet')['type'].value_counts()

====  =========================================================================================================================================================  =======
  ..  type                                                                                                                                                         count
====  =========================================================================================================================================================  =======
   0  string                                                                                                                                                         215
   1  double                                                                                                                                                          22
   2  list<element: string>                                                                                                                                            5
   3  int64                                                                                                                                                            4
   4  list<element: struct<@context: string, @type: string, position: int64, url: string>>                                                                             2
   5  list<element: struct<@context: string, @type: string, contentUrl: string, creditText: string, url: string>>                                                      2
   6  list<element: struct<@context: string, @type: string, caption: string, contentUrl: string, creditText: string, height: int64, url: string, width: int64>>        1
   7  timestamp[ns]                                                                                                                                                    1
   8  list<element: struct<@context: string, @type: string, item: string, name: string, position: int64>>                                                              1
   9  list<element: struct<@context: string, @type: string, name: string, url: string>>                                                                                1
====  =========================================================================================================================================================  =======


Module functions
----------------
"""

import re

import pandas as pd
import pyarrow.parquet as pq

__all__ = [
    "images",
    "links",
    "redirects",
    "jl_subset",
    "jl_to_parquet",
    "parquet_columns",
]


def redirects(crawldf):
    """Create a tidy DataFrame for the redirects in `crawldf` with the columns:

    - url: All the URLs in the redirect (chain).
    - status: The status code of each URL.
    - type: "requested", "inermediate", or "crawled".
    - order: 1, 2, 3... up to the number of urls in the redirect chain.
    - redirect_times: The number of redirects in the chain (URLs in the chain minus one).

    Parameters
    ----------
    crawldf : pandas.DataFrame
        A DataFrame of an advertools crawl file

    Examples
    --------
    >>> import advertools as adv
    >>> import pandas as pd
    >>> crawldf = pd.read_json('output_file.jl', lines=True)
    >>> redirect_df = adv.crawlytics.redirects(crawldf)
    >>> redirect_df

    ====  =========================================================  ========  =======  =========  ==================  ================
      ..  url                                                          status    order  type         download_latency    redirect_times
    ====  =========================================================  ========  =======  =========  ==================  ================
       0  https://nytimes.com                                             301        1  requested           0.220263                  1
       0  https://www.nytimes.com/                                        200        2  crawled             0.220263                  1
      26  https://www.nytimes.com/privacy/privacy-policy                  301        1  requested           0.079844                  1
      26  https://help.nytimes.com/hc/en-us/articles/10940941449492       403        2  crawled             0.079844                  1
     105  https://www.nytimes.com/es/privacy/privacy-policy               301        1  requested           0.0630789                 1
     105  https://help.nytimes.com/hc/en-us/articles/13537530305428       403        2  crawled             0.0630789                 1
    ====  =========================================================  ========  =======  =========  ==================  ================
    """
    if "redirect_urls" not in crawldf.columns:
        return pd.DataFrame()
    if "redirect_urls" in crawldf:
        if crawldf["redirect_urls"].isna().all():
            return pd.DataFrame()
    redirect_df = crawldf[
        ["url", "status", "redirect_urls", "redirect_reasons"]
    ].dropna(subset=["redirect_urls", "redirect_reasons"])
    redirect_df["redirect_urls"] = (
        redirect_df["redirect_urls"].astype(str).str.split("@@")
    )
    redirect_df["redirect_reasons"] = (
        redirect_df["redirect_reasons"].astype(str).str.split("@@")
    )
    for url, redirect_urls in redirect_df[["url", "redirect_urls"]].values:
        redirect_urls.append(url)
    for status, redirect_reasons in redirect_df[["status", "redirect_reasons"]].values:
        redirect_reasons.append(status)
    redirect_df["order"] = [
        list(range(1, len(x) + 1)) for x in redirect_df["redirect_reasons"]
    ]
    redirect_df["type"] = [
        [
            (
                "requested"
                if o == min(order)
                else "crawled" if o == max(order) else "intermediate"
            )
            for o in order
        ]
        for order in redirect_df["order"]
    ]
    redirect_df.columns = ["NA1", "NA2", "url", "status", "order", "type"]
    exploded = redirect_df[["url", "status", "order", "type"]].apply(pd.Series.explode)
    final_df = pd.merge(
        exploded,
        crawldf[["download_latency", "redirect_times"]],
        left_index=True,
        right_index=True,
    )
    final_df["redirect_times"] = final_df["redirect_times"].astype(int)
    return final_df


def links(crawldf, internal_url_regex=None):
    """Summarize links from a crawl DataFrame.

    Parameters
    ----------
    crawldf : DataFrame
        A DataFrame of a website crawled with advertools.
    internal_url_regex : str
        A regular expression for identifying if a link is internal or not.
        For example if your website is example.com, this would be "example.com".

    Returns
    -------
    link_df : pandas.DataFrame

    Examples
    --------
    >>> import advertools as adv
    >>> import pandas as pd
    >>> crawldf = pd.read_json('output_file.jl', lines=True)
    >>> link_df = adv.crawlytics.links(crawldf)
    >>> link_df

    ====  ===========================================================  ========================================================================  ==================  ==========  ==========
      ..  url                                                          link                                                                      text                nofollow    internal
    ====  ===========================================================  ========================================================================  ==================  ==========  ==========
       0  https://www.nytimes.com/                                     https://www.nytimes.com/#site-content                                     Skip to content     False       True
       0  https://www.nytimes.com/                                     https://www.nytimes.com/#site-index                                       Skip to site index  False       True
       0  https://www.nytimes.com/                                     https://www.nytimes.com/#after-dfp-ad-top                                 SKIP ADVERTISEMENT  False       True
       1  https://www.nytimes.com/newsletters/morning-briefing-europe  https://www.nytimes.com/newsletters/morning-briefing-europe#site-content  Skip to content     False       True
       1  https://www.nytimes.com/newsletters/morning-briefing-europe  https://www.nytimes.com/newsletters/morning-briefing-europe#site-index    Skip to site index  False       True
       1  https://www.nytimes.com/newsletters/morning-briefing-europe  https://www.nytimes.com/                                                                      False       True
       2  https://www.nytimes.com/newsletters/australia-letter         https://www.nytimes.com/newsletters/australia-letter#site-content         Skip to content     False       True
       2  https://www.nytimes.com/newsletters/australia-letter         https://www.nytimes.com/newsletters/australia-letter#site-index           Skip to site index  False       True
       2  https://www.nytimes.com/newsletters/australia-letter         https://www.nytimes.com/                                                                      False       True
    ====  ===========================================================  ========================================================================  ==================  ==========  ==========
    """
    if "links_url" not in crawldf:
        return pd.DataFrame()
    link_df = pd.merge(
        crawldf[["url"]],
        crawldf.filter(regex="^links_").apply(lambda s: s.str.split("@@").explode()),
        left_index=True,
        right_index=True,
    )
    link_df["links_nofollow"] = link_df["links_nofollow"].replace(
        {"True": True, "False": False}
    )
    if internal_url_regex is not None:
        link_df["internal"] = (
            link_df["links_url"].fillna("").str.contains(internal_url_regex, regex=True)
        )
    link_df = link_df.rename(
        columns={
            "links_url": "link",
            "links_text": "text",
            "links_nofollow": "nofollow",
        }
    )
    return link_df


def images(crawldf):
    """Summarize crawled images from a crawl DataFrame.

    Parameters
    ----------
    crawldf : pandas.DataFrame
        A crawl DataFrame as a result of the advertools.crawl function.

    Returns
    -------
    img_summary : pandas.DataFrame
        A DataFrame containing all available img tags and their attributes mapped to
        their respective URLs where each image data is represented with a separate row.


    Examples
    --------
    >>> import advertools as adv
    >>> import pandas as pd
    >>> crawldf = pd.read_json('output_file.jl', lines=True)
    >>> image_df = adv.crawlytics.images(crawldf)
    >>> image_df

    ====  ===========================================================  ==================================================================================================================================================================================  ================  =============  ================================  ==============  ===========  ============  ============
      ..  url                                                          img_src                                                                                                                                                                             img_alt           img_loading    img_sizes                         img_decoding      img_width    img_height    img_border
    ====  ===========================================================  ==================================================================================================================================================================================  ================  =============  ================================  ==============  ===========  ============  ============
       0  https://www.nytimes.com/                                     /vi-assets/static-assets/icon-the-morning_144x144-b12a6923b6ad9102b766352261b1a847.webp                                                                                             The Morning Logo                 nan                               nan                     nan           nan           nan
       0  https://www.nytimes.com/                                     /vi-assets/static-assets/icon-the-upshot_144x144-0b1553ff703bbd07ac8fe73e6d215888.webp                                                                                              The Upshot Logo                  nan                               nan                     nan           nan           nan
       0  https://www.nytimes.com/                                     https://static01.nyt.com/images/2017/01/29/podcasts/the-daily-album-art/the-daily-album-art-square320-v5.jpg?quality=75&auto=webp&disable=upscale                                   The Daily Logo                   nan                               nan                     nan           nan           nan
       1  https://www.nytimes.com/newsletters/morning-briefing-europe  https://static.nytimes.com/email-images/NYT-Newsletters-Europe-Icon-500px.jpg                                                                                                       morning briefing  nan            nan                               nan                     nan           nan           nan
       2  https://www.nytimes.com/newsletters/australia-letter         https://static.nytimes.com/email-images/NYT-Newsletters-AustraliaLetter-Icon-500px.jpg                                                                                              australia-letter  nan            nan                               nan                     nan           nan           nan
       3  https://www.nytimes.com/newsletters/the-interpreter          https://static.nytimes.com/email-images/NYT-Newsletters-SONL-TheInterpreter-Icon-500px.jpg                                                                                          the interpreter   nan            nan                               nan                     nan           nan           nan
       4  https://www.nytimes.com/section/world/middleeast             https://static01.nyt.com/images/2024/01/25/multimedia/25israel-1-hbcz/25israel-1-hbcz-thumbWide.jpg?quality=75&auto=webp&disable=upscale                                                              nan            (min-width: 1024px) 205px, 150px  async                   150           100           nan
       4  https://www.nytimes.com/section/world/middleeast             https://static01.nyt.com/images/2024/01/25/multimedia/25israel-hamas-icj-case-explain-wjth/25israel-hamas-icj-case-explain-wjth-thumbWide.jpg?quality=75&auto=webp&disable=upscale                    nan            (min-width: 1024px) 205px, 150px  async                   150           100           nan
       4  https://www.nytimes.com/section/world/middleeast             https://static01.nyt.com/images/2024/01/25/multimedia/25israel-hamas-qatar-israel-ctbv/25israel-hamas-qatar-israel-ctbv-thumbWide.jpg?quality=75&auto=webp&disable=upscale                            nan            (min-width: 1024px) 205px, 150px  async                   150           100           nan
    ====  ===========================================================  ==================================================================================================================================================================================  ================  =============  ================================  ==============  ===========  ============  ============
    """
    dfs = []
    img_df = crawldf.filter(regex="^url$|img_")
    for index, row in img_df.iterrows():
        notna = row.dropna().index
        if len(notna) == 1:
            temp = pd.DataFrame({"url": row["url"]}, index=[index])
        else:
            temp = (
                row.to_frame()
                .T.set_index("url")
                .apply(lambda s: s.str.split("@@"))
                .explode(notna.tolist()[1:])
            )
            temp = temp.reset_index()
            temp.index = [index for i in range(len(temp))]
        dfs.append(temp)
    final_df = pd.concat(dfs)
    return final_df


def jl_subset(filepath, columns=None, regex=None, chunksize=500):
    """Read a jl file extracting selected `columns` and/or columns matching `regex`.

    Parameters
    ----------
    filepath : str
      The path of the .jl (jsonlines) file to read.
    columns : list
      An optional list of column names that you want to read.
    regex : str
      An optional regular expression of the pattern of columns to read.
    chunksize : int
      How many rows to read per chunk.

    Examples
    --------

    >>> import advertools as adv

    Read only the columns "url" and "meta_desc":

    >>> adv.crawlytics.jl_subset('output_file.jl', columns=['url', 'meta_desc'])

    Read columns matching the regex "jsonld":

    >>> adv.crawlytics.jl_subset('output_file.jl', regex='jsonld')

    Read the columns "url" and "meta_desc" as well as columns matching "jsonld":

    >>> adv.crawlytics.jl_subset('output_file.jl', columns=['url', 'meta_desc'], regex='jsonld')

    Returns
    -------
    df_subset : pandas.DataFrame
      A DataFrame containing the list of `columns` and/or columns matching `regex`.
    """
    if columns is None and regex is None:
        raise ValueError("Please supply either a list of columns or a regex.")
    if columns is not None:
        col_regex = "^" + "$|^".join(columns) + "$"
    else:
        col_regex = None
    if (columns is not None) and (regex is not None):
        full_regex = "|".join([col_regex, regex])
    else:
        full_regex = col_regex or regex
    dfs = []
    for chunk in pd.read_json(filepath, lines=True, chunksize=chunksize):
        chunk_subset = chunk.filter(regex=full_regex)
        dfs.append(chunk_subset)
    final_df = pd.concat(dfs, ignore_index=True)
    return final_df


def jl_to_parquet(jl_filepath, parquet_filepath):
    """Convert a jsonlines crawl file to the parquet format.

    Parameters
    ----------
    jl_filepath : str
      The path of an existing .jl file.
    parquet_fileapth : str
      The pather where you want the new file to be saved.

    Examples
    --------
    >>> import advertools as adv
    >>> adv.crawlytics.jl_to_parquet("output_file.jl", "output_file.parquet")
    """
    status = "not done"
    crawldf = pd.read_json(jl_filepath, lines=True)
    while status == "not done":
        try:
            crawldf.to_parquet(parquet_filepath, index=False, version="2.6")
            status = "done"
        except Exception as e:
            error = e.args[-1]
            column = re.findall(r"column (\S+)", error)
            print(f"converting to string: {column[0]}")
            crawldf[column[0]] = crawldf[column[0]].astype(str).replace("nan", pd.NA)


def parquet_columns(filepath):
    """Get column names and datatypes of a parquet file.

    Parameters
    ----------
    filepath : str
      The path of the file that you want to get columns names and types.

    Returns
    -------
    columns_types : pandas.DataFrame
      A DataFrame with two columns "column" and "type".
    """
    pqdataset = pq.ParquetDataset(filepath)
    columns_df = pd.DataFrame(
        zip(pqdataset.schema.names, pqdataset.schema.types), columns=["column", "type"]
    )
    return columns_df
